/*
    SPDX-License-Identifier: GPL-3.0

    Copyright (C) 2020-2025  Jevgenijs Protopopovs

    This file is part of Kefir project.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef KEFIR_IR_ASSEMBLY_H_
#define KEFIR_IR_ASSEMBLY_H_

#include "kefir/core/basic-types.h"
#include "kefir/core/list.h"
#include "kefir/core/hashtree.h"
#include "kefir/core/mem.h"
#include "kefir/core/string_pool.h"
#include "kefir/ir/type.h"

typedef enum kefir_ir_inline_assembly_parameter_class {
    KEFIR_IR_INLINE_ASSEMBLY_PARAMETER_IMMEDIATE,
    KEFIR_IR_INLINE_ASSEMBLY_PARAMETER_READ,
    KEFIR_IR_INLINE_ASSEMBLY_PARAMETER_LOAD,
    KEFIR_IR_INLINE_ASSEMBLY_PARAMETER_STORE,
    KEFIR_IR_INLINE_ASSEMBLY_PARAMETER_READ_STORE,
    KEFIR_IR_INLINE_ASSEMBLY_PARAMETER_LOAD_STORE
} kefir_ir_inline_assembly_parameter_class_t;

typedef struct kefir_ir_inline_assembly_parameter_constraints {
    kefir_bool_t general_purpose_register;
    kefir_bool_t floating_point_register;
    kefir_bool_t x87_stack;
    kefir_bool_t memory_location;
    kefir_bool_t immediate;
    kefir_bool_t strict_immediate;

    const char *explicit_register;
} kefir_ir_inline_assembly_parameter_constraints_t;

typedef enum kefir_ir_inline_assembly_immediate_type {
    KEFIR_IR_INLINE_ASSEMBLY_IMMEDIATE_IDENTIFIER_BASED,
    KEFIR_IR_INLINE_ASSEMBLY_IMMEDIATE_LITERAL_BASED
} kefir_ir_inline_assembly_immediate_type_t;

typedef struct kefir_ir_inline_assembly_parameter {
    kefir_id_t parameter_id;
    struct kefir_list identifiers;
    kefir_ir_inline_assembly_parameter_class_t klass;
    struct {
        const struct kefir_ir_type *type;
        kefir_id_t type_id;
        kefir_size_t index;
    } read_type;
    struct {
        const struct kefir_ir_type *type;
        kefir_id_t type_id;
        kefir_size_t index;
    } type;
    struct kefir_ir_inline_assembly_parameter_constraints constraints;
    union {
        struct {
            kefir_size_t read_index;
            kefir_size_t load_store_index;
        };
        struct {
            kefir_ir_inline_assembly_immediate_type_t immediate_type;
            union {
                const char *immediate_identifier_base;
                kefir_id_t immediate_literal_base;
            };
            kefir_int64_t immediate_value;
        };
    };
} kefir_ir_inline_assembly_parameter_t;

typedef struct kefir_ir_inline_assembly_jump_target {
    kefir_id_t uid;
    struct kefir_list identifiers;
    const char *target_function;
    kefir_size_t target;
} kefir_ir_inline_assembly_jump_target_t;

typedef struct kefir_ir_inline_assembly {
    kefir_id_t id;
    const char *template;
    struct kefir_hashtree parameters;
    struct kefir_list parameter_list;
    struct kefir_hashtree clobbers;
    struct kefir_hashtree jump_targets;
    struct kefir_list jump_target_list;
    kefir_id_t next_jump_target_id;
} kefir_ir_inline_assembly_t;

struct kefir_ir_inline_assembly *kefir_ir_inline_assembly_alloc(struct kefir_mem *, struct kefir_string_pool *,
                                                                kefir_id_t, const char *);
kefir_result_t kefir_ir_inline_assembly_free(struct kefir_mem *, struct kefir_ir_inline_assembly *);
kefir_result_t kefir_ir_inline_assembly_add_parameter(struct kefir_mem *, struct kefir_string_pool *,
                                                      struct kefir_ir_inline_assembly *, const char *,
                                                      kefir_ir_inline_assembly_parameter_class_t,
                                                      const struct kefir_ir_inline_assembly_parameter_constraints *,
                                                      const struct kefir_ir_type *, kefir_id_t, kefir_size_t,
                                                      kefir_size_t, struct kefir_ir_inline_assembly_parameter **);
kefir_result_t kefir_ir_inline_assembly_add_parameter_alias(struct kefir_mem *, struct kefir_string_pool *,
                                                            struct kefir_ir_inline_assembly *,
                                                            struct kefir_ir_inline_assembly_parameter *, const char *);
kefir_result_t kefir_ir_inline_assembly_add_immediate_parameter(struct kefir_mem *, struct kefir_string_pool *,
                                                                struct kefir_ir_inline_assembly *, const char *,
                                                                const struct kefir_ir_type *, kefir_id_t, kefir_size_t,
                                                                kefir_ir_inline_assembly_immediate_type_t, const char *,
                                                                kefir_id_t, kefir_int64_t,
                                                                struct kefir_ir_inline_assembly_parameter **);
kefir_result_t kefir_ir_inline_assembly_parameter_read_from(struct kefir_mem *, struct kefir_ir_inline_assembly *,
                                                            struct kefir_ir_inline_assembly_parameter *,
                                                            const struct kefir_ir_type *, kefir_id_t, kefir_size_t,
                                                            kefir_size_t);
kefir_result_t kefir_ir_inline_assembly_resolve_parameter(struct kefir_ir_inline_assembly *, const char *,
                                                          struct kefir_ir_inline_assembly_parameter **);
kefir_result_t kefir_ir_inline_assembly_add_clobber(struct kefir_mem *, struct kefir_string_pool *,
                                                    struct kefir_ir_inline_assembly *, const char *);
kefir_result_t kefir_ir_inline_assembly_add_jump_target(struct kefir_mem *, struct kefir_string_pool *,
                                                        struct kefir_ir_inline_assembly *, const char *, const char *,
                                                        kefir_size_t, struct kefir_ir_inline_assembly_jump_target **);
kefir_result_t kefir_ir_inline_assembly_add_jump_target_alias(struct kefir_mem *, struct kefir_string_pool *,
                                                              struct kefir_ir_inline_assembly *,
                                                              struct kefir_ir_inline_assembly_jump_target *,
                                                              const char *);

#endif
