/*
    SPDX-License-Identifier: GPL-3.0

    Copyright (C) 2020-2025  Jevgenijs Protopopovs

    This file is part of Kefir project.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "kefir/ast-translator/translator.h"
#include "kefir/ast-translator/scope/local_scope_layout.h"
#include "kefir/ast-translator/flow_control.h"
#include "kefir/ast-translator/lvalue.h"
#include "kefir/ast-translator/value.h"
#include "kefir/ast-translator/scope/translator.h"
#include "kefir/ast-translator/typeconv.h"
#include "kefir/ast-translator/temporaries.h"
#include "kefir/ast-translator/debug/translator.h"
#include "kefir/ast/type_conv.h"
#include "kefir/ast/downcast.h"
#include "kefir/core/util.h"
#include "kefir/core/error.h"
#include "kefir/ast-translator/function_definition.h"
#include "kefir/ast/runtime.h"
#include <stdio.h>

static kefir_result_t init_function_declaration(struct kefir_mem *mem, struct kefir_ast_translator_context *context,
                                                struct kefir_ast_function_definition *function,
                                                struct kefir_ast_translator_function_context *args) {
    struct kefir_ast_declarator_identifier *decl_identifier = NULL;
    REQUIRE_OK(kefir_ast_declarator_unpack_identifier(function->declarator, &decl_identifier));
    REQUIRE(decl_identifier != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Expected function definition to have valid identifier"));

    const char *identifier = decl_identifier->identifier;

    char identifier_buf[1024];
    const struct kefir_ast_scoped_identifier *scoped_id = function->base.properties.function_definition.scoped_id;
    if (scoped_id->function.flags.gnu_inline &&
        scoped_id->function.storage == KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_EXTERN &&
        kefir_ast_function_specifier_is_inline(scoped_id->function.specifier) &&
        !scoped_id->function.inline_definition && scoped_id->function.asm_label == NULL) {
        snprintf(identifier_buf, sizeof(identifier_buf) - 1, KEFIR_AST_TRANSLATOR_GNU_INLINE_FUNCTION_IDENTIFIER,
                 identifier);
        identifier = kefir_string_pool_insert(mem, context->ast_context->symbols, identifier_buf, NULL);
        REQUIRE(identifier != NULL,
                KEFIR_SET_ERROR(KEFIR_OBJALLOC_FAILURE,
                                "Failed to insert generated function identifier into symbol table"));
    }

    const struct kefir_ast_declarator_function *decl_func = NULL;
    REQUIRE_OK(kefir_ast_declarator_unpack_function(function->declarator, &decl_func));
    REQUIRE(decl_func != NULL,
            KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected function definition to have function declarator"));
    switch (function->base.properties.type->function_type.mode) {
        case KEFIR_AST_FUNCTION_TYPE_PARAMETERS:
        case KEFIR_AST_FUNCTION_TYPE_PARAM_EMPTY:
            REQUIRE_OK(kefir_ast_translator_function_declaration_init(
                mem, context->ast_context, context->environment, context->ast_context->type_bundle,
                context->ast_context->type_traits, context->module, identifier,
                function->base.properties.function_definition.scoped_id->type, NULL, &args->function_declaration,
                &function->base.source_location));
            break;

        case KEFIR_AST_FUNCTION_TYPE_PARAM_IDENTIFIERS: {
            struct kefir_list declaration_list;
            struct kefir_hashtree declarations;

            kefir_result_t res = kefir_hashtree_init(&declarations, &kefir_hashtree_str_ops);
            for (const struct kefir_list_entry *iter = kefir_list_head(&function->declarations);
                 res == KEFIR_OK && iter != NULL; kefir_list_next(&iter)) {

                ASSIGN_DECL_CAST(struct kefir_ast_node_base *, decl_node, iter->value);
                struct kefir_ast_declaration *decl_list = NULL;
                REQUIRE_MATCH(&res, kefir_ast_downcast_declaration(decl_node, &decl_list, false),
                              KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Expected AST declaration"));

                for (const struct kefir_list_entry *decl_iter = kefir_list_head(&decl_list->init_declarators);
                     decl_iter != NULL && res == KEFIR_OK; kefir_list_next(&decl_iter)) {
                    ASSIGN_DECL_CAST(struct kefir_ast_node_base *, decl, decl_iter->value);
                    REQUIRE_CHAIN(&res, kefir_hashtree_insert(
                                            mem, &declarations,
                                            (kefir_hashtree_key_t) decl->properties.declaration_props.identifier,
                                            (kefir_hashtree_value_t) decl));
                }
            }

            REQUIRE_CHAIN(&res, kefir_list_init(&declaration_list));
            REQUIRE_ELSE(res == KEFIR_OK, {
                kefir_hashtree_free(mem, &declarations);
                return res;
            });

            for (const struct kefir_list_entry *iter = kefir_list_head(&decl_func->parameters);
                 res == KEFIR_OK && iter != NULL; kefir_list_next(&iter)) {

                struct kefir_hashtree_node *tree_node = NULL;
                ASSIGN_DECL_CAST(struct kefir_ast_node_base *, param, iter->value);
                REQUIRE_CHAIN_SET(&res,
                                  param->properties.category == KEFIR_AST_NODE_CATEGORY_EXPRESSION &&
                                      param->properties.expression_props.identifier != NULL,
                                  KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected parameter to be AST identifier"));
                REQUIRE_CHAIN(&res,
                              kefir_hashtree_at(&declarations,
                                                (kefir_hashtree_key_t) param->properties.expression_props.identifier,
                                                &tree_node));

                REQUIRE_CHAIN(&res, kefir_list_insert_after(mem, &declaration_list, kefir_list_tail(&declaration_list),
                                                            (struct kefir_ast_node_base *) tree_node->value));
            }

            REQUIRE_ELSE(res == KEFIR_OK, {
                kefir_list_free(mem, &declaration_list);
                kefir_hashtree_free(mem, &declarations);
                return res;
            });

            res = kefir_ast_translator_function_declaration_init(
                mem, context->ast_context, context->environment, context->ast_context->type_bundle,
                context->ast_context->type_traits, context->module, identifier, function->base.properties.type,
                &declaration_list, &args->function_declaration, &function->base.source_location);
            REQUIRE_ELSE(res == KEFIR_OK, {
                kefir_list_free(mem, &declaration_list);
                kefir_hashtree_free(mem, &declarations);
                return res;
            });

            res = kefir_list_free(mem, &declaration_list);
            REQUIRE_ELSE(res == KEFIR_OK, {
                if (args->function_declaration != NULL) {
                    kefir_ast_translator_function_declaration_free(mem, args->function_declaration);
                }
                kefir_hashtree_free(mem, &declarations);
                return res;
            });

            res = kefir_hashtree_free(mem, &declarations);
            REQUIRE_ELSE(res == KEFIR_OK, {
                if (args->function_declaration != NULL) {
                    kefir_ast_translator_function_declaration_free(mem, args->function_declaration);
                }
                return res;
            });
        } break;
    }

    return KEFIR_OK;
}

static kefir_result_t free_function_declaration(struct kefir_mem *mem,
                                                struct kefir_ast_translator_function_context *args) {
    if (args->function_declaration != NULL) {
        REQUIRE_OK(kefir_ast_translator_function_declaration_free(mem, args->function_declaration));
    }
    args->function_declaration = NULL;
    return KEFIR_OK;
}

kefir_result_t kefir_ast_translator_function_context_init(struct kefir_mem *mem,
                                                          struct kefir_ast_translator_context *context,
                                                          struct kefir_ast_function_definition *function,
                                                          struct kefir_ast_translator_function_context *ctx) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(context != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AST translator context"));
    REQUIRE(function != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AST function definition"));
    REQUIRE(ctx != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to AST translator function context"));

    REQUIRE_OK(init_function_declaration(mem, context, function, ctx));

    ctx->function_definition = function;
    ctx->module = context->module;
    ctx->local_context = function->base.properties.function_definition.scoped_id->function.local_context;
    kefir_result_t res = kefir_ast_translator_context_init_local(mem, &ctx->local_translator_context,
                                                                 &ctx->local_context->context, NULL, context);
    REQUIRE_ELSE(res == KEFIR_OK, {
        free_function_declaration(mem, ctx);
        return res;
    });

    res = kefir_ast_translator_local_scope_layout_init(mem, context->module, context->global_scope_layout,
                                                       &ctx->local_scope_layout);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_ast_translator_context_free(mem, &ctx->local_translator_context);
        free_function_declaration(mem, ctx);
        return res;
    });

    res = kefir_ast_translator_build_local_scope_layout(
        mem, ctx->local_context, ctx->local_translator_context.environment, ctx->local_translator_context.module,
        &ctx->local_scope_layout, context->debug_entries);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_ast_translator_local_scope_layout_free(mem, &ctx->local_scope_layout);
        kefir_ast_translator_context_free(mem, &ctx->local_translator_context);
        free_function_declaration(mem, ctx);
        return res;
    });

    res = kefir_ast_translator_flow_control_tree_init(mem, ctx->local_context->context.flow_control_tree);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_ast_translator_local_scope_layout_free(mem, &ctx->local_scope_layout);
        kefir_ast_translator_context_free(mem, &ctx->local_translator_context);
        free_function_declaration(mem, ctx);
        return res;
    });

    ctx->ir_func = kefir_ir_module_new_function(mem, context->module, ctx->function_declaration->ir_function_decl,
                                                ctx->local_scope_layout.local_layout_id, 0);
    REQUIRE_ELSE(ctx->ir_func != NULL, {
        kefir_ast_translator_local_scope_layout_free(mem, &ctx->local_scope_layout);
        kefir_ast_translator_context_free(mem, &ctx->local_translator_context);
        free_function_declaration(mem, ctx);
        return KEFIR_SET_ERROR(KEFIR_OBJALLOC_FAILURE, "Failed to allocate IR function");
    });

    res = kefir_irbuilder_block_init(mem, &ctx->builder, &ctx->ir_func->body);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_ast_translator_local_scope_layout_free(mem, &ctx->local_scope_layout);
        kefir_ast_translator_context_free(mem, &ctx->local_translator_context);
        free_function_declaration(mem, ctx);
        return res;
    });

    ctx->ir_func->flags.constructor =
        function->base.properties.function_definition.scoped_id->function.flags.constructor;
    ctx->ir_func->flags.destructor = function->base.properties.function_definition.scoped_id->function.flags.destructor;

    ctx->local_translator_context.function_debug_info = &ctx->ir_func->debug_info;

    return KEFIR_OK;
}

kefir_result_t kefir_ast_translator_function_context_free(struct kefir_mem *mem,
                                                          struct kefir_ast_translator_function_context *ctx) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(ctx != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AST translator function context"));

    kefir_result_t res = KEFIR_IRBUILDER_BLOCK_FREE(&ctx->builder);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_ast_translator_local_scope_layout_free(mem, &ctx->local_scope_layout);
        kefir_ast_translator_context_free(mem, &ctx->local_translator_context);
        free_function_declaration(mem, ctx);
        return res;
    });

    res = kefir_ast_translator_local_scope_layout_free(mem, &ctx->local_scope_layout);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_ast_translator_context_free(mem, &ctx->local_translator_context);
        free_function_declaration(mem, ctx);
        return res;
    });

    res = kefir_ast_translator_context_free(mem, &ctx->local_translator_context);
    REQUIRE_ELSE(res == KEFIR_OK, {
        free_function_declaration(mem, ctx);
        return res;
    });

    REQUIRE_OK(free_function_declaration(mem, ctx));
    ctx->function_definition = NULL;
    return KEFIR_OK;
}

static kefir_result_t xchg_param_address(struct kefir_irbuilder_block *builder) {
    REQUIRE_OK(KEFIR_IRBUILDER_BLOCK_APPENDI64(builder, KEFIR_IROPCODE_XCHG, 1));
    return KEFIR_OK;
}

struct vl_modified_param {
    struct kefir_mem *mem;
    struct kefir_irbuilder_block *builder;
    struct kefir_ast_translator_context *context;
};

static kefir_result_t translate_variably_modified(const struct kefir_ast_node_base *node, void *payload) {
    REQUIRE(node != NULL, KEFIR_OK);
    REQUIRE(payload != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid payload"));
    ASSIGN_DECL_CAST(struct vl_modified_param *, param, payload);

    REQUIRE_OK(kefir_ast_translate_expression(param->mem, node, param->builder, param->context));
    REQUIRE_OK(KEFIR_IRBUILDER_BLOCK_APPENDU64(param->builder, KEFIR_IROPCODE_POP, 0));
    return KEFIR_OK;
}

static kefir_result_t generate_debug_info(struct kefir_mem *mem,
                                          struct kefir_ast_translator_function_context *function_context,
                                          kefir_ir_debug_entry_id_t subprogram_entry_id) {
    REQUIRE_OK(kefir_ir_debug_entry_add_attribute(mem, &function_context->module->debug_info.entries,
                                                  &function_context->module->symbols, subprogram_entry_id,
                                                  &KEFIR_IR_DEBUG_ENTRY_ATTR_NAME(function_context->ir_func->name)));

    kefir_ir_debug_entry_id_t subprogram_return_type_id;
    REQUIRE_OK(kefir_ast_translate_debug_type(
        mem, &function_context->local_context->global->context, function_context->local_translator_context.environment,
        function_context->module, function_context->local_translator_context.debug_entries,
        function_context->function_declaration->function_type->function_type.return_type, &subprogram_return_type_id));
    REQUIRE_OK(kefir_ir_debug_entry_add_attribute(mem, &function_context->module->debug_info.entries,
                                                  &function_context->module->symbols, subprogram_entry_id,
                                                  &KEFIR_IR_DEBUG_ENTRY_ATTR_TYPE(subprogram_return_type_id)));

    const struct kefir_ast_scoped_identifier *scoped_id =
        function_context->function_definition->base.properties.function_definition.scoped_id;
    REQUIRE_OK(kefir_ir_debug_entry_add_attribute(
        mem, &function_context->module->debug_info.entries, &function_context->module->symbols, subprogram_entry_id,
        &KEFIR_IR_DEBUG_ENTRY_ATTR_EXTERNAL(scoped_id->function.storage != KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_STATIC)));
    REQUIRE_OK(kefir_ir_debug_entry_add_attribute(
        mem, &function_context->module->debug_info.entries, &function_context->module->symbols, subprogram_entry_id,
        &KEFIR_IR_DEBUG_ENTRY_ATTR_DECLARATION(!scoped_id->function.defined)));

    if (function_context->function_definition->base.source_location.source != NULL) {
        REQUIRE_OK(kefir_ir_debug_entry_add_attribute(
            mem, &function_context->module->debug_info.entries, &function_context->module->symbols, subprogram_entry_id,
            &KEFIR_IR_DEBUG_ENTRY_ATTR_SOURCE_LOCATION(
                function_context->function_definition->base.source_location.source)));
        REQUIRE_OK(kefir_ir_debug_entry_add_attribute(
            mem, &function_context->module->debug_info.entries, &function_context->module->symbols, subprogram_entry_id,
            &KEFIR_IR_DEBUG_ENTRY_ATTR_SOURCE_LOCATION_LINE(
                function_context->function_definition->base.source_location.line)));
        REQUIRE_OK(kefir_ir_debug_entry_add_attribute(
            mem, &function_context->module->debug_info.entries, &function_context->module->symbols, subprogram_entry_id,
            &KEFIR_IR_DEBUG_ENTRY_ATTR_SOURCE_LOCATION_COLUMN(
                function_context->function_definition->base.source_location.column)));
    }
    return KEFIR_OK;
}

static kefir_result_t generate_parameter_debug_info(struct kefir_mem *mem,
                                                    struct kefir_ast_translator_function_context *function_context,
                                                    kefir_ir_debug_entry_id_t subprogram_entry_id, const char *name,
                                                    const struct kefir_ast_type *type, kefir_size_t parameter_index) {
    kefir_ir_debug_entry_id_t parameter_entry_id, parameter_type_id;
    REQUIRE_OK(kefir_ir_debug_entry_new_child(mem, &function_context->module->debug_info.entries, subprogram_entry_id,
                                              KEFIR_IR_DEBUG_ENTRY_FUNCTION_PARAMETER, &parameter_entry_id));

    if (name != NULL) {
        REQUIRE_OK(kefir_ir_debug_entry_add_attribute(mem, &function_context->module->debug_info.entries,
                                                      &function_context->module->symbols, parameter_entry_id,
                                                      &KEFIR_IR_DEBUG_ENTRY_ATTR_NAME(name)));
    }

    REQUIRE_OK(kefir_ast_translate_debug_type(
        mem, &function_context->local_context->global->context, function_context->local_translator_context.environment,
        function_context->module, function_context->local_translator_context.debug_entries, type, &parameter_type_id));
    REQUIRE_OK(kefir_ir_debug_entry_add_attribute(mem, &function_context->module->debug_info.entries,
                                                  &function_context->module->symbols, parameter_entry_id,
                                                  &KEFIR_IR_DEBUG_ENTRY_ATTR_TYPE(parameter_type_id)));

    REQUIRE_OK(kefir_ir_debug_entry_add_attribute(mem, &function_context->module->debug_info.entries,
                                                  &function_context->module->symbols, parameter_entry_id,
                                                  &KEFIR_IR_DEBUG_ENTRY_ATTR_PARAMETER(parameter_index)));
    return KEFIR_OK;
}

kefir_result_t kefir_ast_translator_function_context_translate(
    struct kefir_mem *mem, struct kefir_ast_translator_function_context *function_context) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(function_context != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AST translator function context"));

    struct kefir_ast_function_definition *function = function_context->function_definition;
    struct kefir_irbuilder_block *builder = &function_context->builder;
    struct kefir_ast_translator_context *context = &function_context->local_translator_context;

    kefir_ir_debug_entry_id_t subprogram_entry_id;
    REQUIRE_OK(kefir_ast_translator_context_push_debug_hierarchy_entry(
        mem, &function_context->local_translator_context, KEFIR_IR_DEBUG_ENTRY_SUBPROGRAM, &subprogram_entry_id));

    if (context->function_debug_info != NULL) {
        context->function_debug_info->subprogram_id = subprogram_entry_id;
        REQUIRE_OK(kefir_ir_function_debug_info_set_source_location(
            mem, context->function_debug_info, &context->module->symbols, &function->base.source_location));
        REQUIRE_OK(generate_debug_info(mem, function_context, subprogram_entry_id));
    }

    const struct kefir_ast_declarator_function *decl_func = NULL;
    REQUIRE_OK(kefir_ast_declarator_unpack_function(function->declarator, &decl_func));
    REQUIRE(decl_func != NULL,
            KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected function definition to have function declarator"));
    kefir_result_t res;
    for (const struct kefir_list_entry *iter = kefir_list_tail(&decl_func->parameters); iter != NULL;
         iter = iter->prev) {

        ASSIGN_DECL_CAST(struct kefir_ast_node_base *, param, iter->value);
        const struct kefir_ast_scoped_identifier *scoped_id = NULL;

        if (param->properties.category == KEFIR_AST_NODE_CATEGORY_DECLARATION) {
            struct kefir_ast_declaration *param_decl = NULL;
            REQUIRE_MATCH_OK(
                &res, kefir_ast_downcast_declaration(param, &param_decl, false),
                KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Expected function parameter to be an AST declaration"));
            REQUIRE(kefir_list_length(&param_decl->init_declarators) == 1,
                    KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Expected function parameter to have exactly one declarator"));

            ASSIGN_DECL_CAST(struct kefir_ast_init_declarator *, init_decl,
                             kefir_list_head(&param_decl->init_declarators)->value);
            struct kefir_ast_declarator_identifier *param_identifier = NULL;
            REQUIRE_OK(kefir_ast_declarator_unpack_identifier(init_decl->declarator, &param_identifier));
            REQUIRE(init_decl->base.properties.type != NULL,
                    KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Function definition parameters shall have definite types"));
            if (init_decl->base.properties.type->tag != KEFIR_AST_TYPE_VOID) {
                if (param_identifier != NULL && param_identifier->identifier != NULL) {
                    scoped_id = init_decl->base.properties.declaration_props.scoped_id;
                    REQUIRE_OK(kefir_ast_translator_object_lvalue(mem, context, builder, param_identifier->identifier,
                                                                  scoped_id));
                    REQUIRE_OK(xchg_param_address(builder));
                    REQUIRE_OK(
                        kefir_ast_translator_store_value(mem, scoped_id->type, context, builder,
                                                         &function_context->function_definition->base.source_location));
                } else {
                    REQUIRE_OK(KEFIR_IRBUILDER_BLOCK_APPENDI64(builder, KEFIR_IROPCODE_POP, 0));
                }
            }
        } else if (param->properties.category == KEFIR_AST_NODE_CATEGORY_EXPRESSION &&
                   param->properties.expression_props.identifier != NULL) {
            scoped_id = param->properties.expression_props.scoped_id;
            REQUIRE_OK(kefir_ast_translator_object_lvalue(mem, context, builder,
                                                          param->properties.expression_props.identifier, scoped_id));
            REQUIRE_OK(xchg_param_address(builder));

            const struct kefir_ast_type *default_promotion =
                kefir_ast_type_function_default_argument_convertion_promotion(
                    mem, context->ast_context->type_bundle, context->ast_context->type_traits, scoped_id->object.type);
            if (KEFIR_AST_TYPE_IS_SCALAR_TYPE(default_promotion)) {
                REQUIRE_OK(kefir_ast_translate_typeconv(mem, context->module, builder,
                                                        context->ast_context->type_traits, default_promotion,
                                                        scoped_id->object.type));
            }

            REQUIRE_OK(kefir_ast_translator_store_value(mem, scoped_id->object.type, context, builder,
                                                        &function_context->function_definition->base.source_location));
        } else {
            return KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR,
                                   "Expected function parameter to be either AST declaration or identifier");
        }
    }

    // Generate debug information for parameters
    if (context->function_debug_info != NULL) {
        kefir_size_t parameter_index = 0;
        for (const struct kefir_list_entry *iter = kefir_list_head(&decl_func->parameters); iter != NULL;
             kefir_list_next(&iter), parameter_index++) {

            ASSIGN_DECL_CAST(struct kefir_ast_node_base *, param, iter->value);
            const struct kefir_ast_scoped_identifier *scoped_id = NULL;

            if (param->properties.category == KEFIR_AST_NODE_CATEGORY_DECLARATION) {
                struct kefir_ast_declaration *param_decl = NULL;
                REQUIRE_MATCH_OK(
                    &res, kefir_ast_downcast_declaration(param, &param_decl, false),
                    KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Expected function parameter to be an AST declaration"));

                ASSIGN_DECL_CAST(struct kefir_ast_init_declarator *, init_decl,
                                 kefir_list_head(&param_decl->init_declarators)->value);
                struct kefir_ast_declarator_identifier *param_identifier = NULL;
                REQUIRE_OK(kefir_ast_declarator_unpack_identifier(init_decl->declarator, &param_identifier));
                if (init_decl->base.properties.type->tag != KEFIR_AST_TYPE_VOID) {
                    if (param_identifier != NULL && param_identifier->identifier != NULL) {
                        scoped_id = init_decl->base.properties.declaration_props.scoped_id;
                        REQUIRE_OK(generate_parameter_debug_info(mem, function_context, subprogram_entry_id,
                                                                 param_identifier->identifier, scoped_id->type,
                                                                 parameter_index));
                    } else {
                        REQUIRE_OK(generate_parameter_debug_info(mem, function_context, subprogram_entry_id, NULL,
                                                                 init_decl->base.properties.type, parameter_index));
                    }
                }
            } else if (param->properties.category == KEFIR_AST_NODE_CATEGORY_EXPRESSION &&
                       param->properties.expression_props.identifier != NULL) {
                scoped_id = param->properties.expression_props.scoped_id;
                REQUIRE_OK(generate_parameter_debug_info(mem, function_context, subprogram_entry_id,
                                                         param->properties.expression_props.identifier,
                                                         scoped_id->object.type, parameter_index));
            } else {
                return KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR,
                                       "Expected function parameter to be either AST declaration or identifier");
            }
        }

        if (function_context->function_declaration->function_type->function_type.ellipsis) {
            kefir_ir_debug_entry_id_t parameter_entry_id;
            REQUIRE_OK(kefir_ir_debug_entry_new_child(mem, &function_context->module->debug_info.entries,
                                                      subprogram_entry_id, KEFIR_IR_DEBUG_ENTRY_FUNCTION_VARARG,
                                                      &parameter_entry_id));
        }
    }

    // Translate variably-modified types
    for (const struct kefir_list_entry *iter = kefir_list_tail(&decl_func->parameters); iter != NULL;
         iter = iter->prev) {

        ASSIGN_DECL_CAST(struct kefir_ast_node_base *, param, iter->value);
        if (param->properties.category == KEFIR_AST_NODE_CATEGORY_DECLARATION) {
            struct kefir_ast_declaration *param_decl = NULL;
            REQUIRE_MATCH_OK(
                &res, kefir_ast_downcast_declaration(param, &param_decl, false),
                KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Expected function parameter to be an AST declaration"));
            REQUIRE(kefir_list_length(&param_decl->init_declarators) == 1,
                    KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Expected function parameter to have exactly one declarator"));

            ASSIGN_DECL_CAST(struct kefir_ast_init_declarator *, init_decl,
                             kefir_list_head(&param_decl->init_declarators)->value);

            REQUIRE_OK(kefir_ast_type_list_variable_modificators(
                init_decl->base.properties.declaration_props.original_type, translate_variably_modified,
                &(struct vl_modified_param){.mem = mem, .context = context, .builder = builder}));
        }
    }

    for (const struct kefir_list_entry *iter = kefir_list_head(&function_context->function_definition->declarations);
         iter != NULL; kefir_list_next(&iter)) {

        ASSIGN_DECL_CAST(struct kefir_ast_node_base *, decl_node, iter->value);
        struct kefir_ast_declaration *decl_list = NULL;
        REQUIRE_MATCH_OK(&res, kefir_ast_downcast_declaration(decl_node, &decl_list, false),
                         KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Expected AST node to be a declaration"));

        for (const struct kefir_list_entry *decl_iter = kefir_list_head(&decl_list->init_declarators);
             decl_iter != NULL; decl_list = kefir_list_next(&decl_iter)) {

            ASSIGN_DECL_CAST(struct kefir_ast_init_declarator *, decl, decl_iter->value);

            REQUIRE_OK(kefir_ast_type_list_variable_modificators(
                decl->base.properties.declaration_props.original_type, translate_variably_modified,
                &(struct vl_modified_param){.mem = mem, .context = context, .builder = builder}));
        }
    }

    const kefir_size_t function_begin_index = KEFIR_IRBUILDER_BLOCK_CURRENT_INDEX(builder);
    REQUIRE_OK(kefir_ir_debug_entry_add_attribute(mem, &context->module->debug_info.entries, &context->module->symbols,
                                                  subprogram_entry_id,
                                                  &KEFIR_IR_DEBUG_ENTRY_ATTR_CODE_BEGIN(function_begin_index)));

    for (const struct kefir_list_entry *iter = kefir_list_head(&function->body->block_items); iter != NULL;
         kefir_list_next(&iter)) {
        ASSIGN_DECL_CAST(struct kefir_ast_node_base *, item, iter->value);

        if (item->properties.category == KEFIR_AST_NODE_CATEGORY_STATEMENT ||
            item->properties.category == KEFIR_AST_NODE_CATEGORY_INLINE_ASSEMBLY) {
            REQUIRE_OK(kefir_ast_translate_statement(mem, item, builder, context));
        } else if (item->properties.category == KEFIR_AST_NODE_CATEGORY_DECLARATION ||
                   item->properties.category == KEFIR_AST_NODE_CATEGORY_INIT_DECLARATOR) {
            REQUIRE_OK(kefir_ast_translate_declaration(mem, item, builder, context));
        } else {
            return KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Unexpected compound statement item");
        }
    }

    const kefir_size_t function_end_index = KEFIR_IRBUILDER_BLOCK_CURRENT_INDEX(builder);
    REQUIRE_OK(kefir_ir_debug_entry_add_attribute(mem, &context->module->debug_info.entries, &context->module->symbols,
                                                  subprogram_entry_id,
                                                  &KEFIR_IR_DEBUG_ENTRY_ATTR_CODE_END(function_end_index)));
    const struct kefir_ast_identifier_flat_scope *associated_ordinary_scope =
        function->body->base.properties.statement_props.flow_control_statement->associated_scopes.ordinary_scope;
    REQUIRE(associated_ordinary_scope != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_STATE,
                            "Expected AST flow control statement to have an associated ordinary scope"));
    REQUIRE_OK(kefir_ast_translator_generate_object_scope_debug_information(
        mem, context->ast_context, context->environment, context->module, context->debug_entries,
        associated_ordinary_scope, subprogram_entry_id, function_begin_index, function_end_index));
    REQUIRE_OK(
        kefir_ast_translator_context_pop_debug_hierarchy_entry(mem, &function_context->local_translator_context));

    return KEFIR_OK;
}

kefir_result_t kefir_ast_translator_function_context_finalize(
    struct kefir_mem *mem, struct kefir_ast_translator_function_context *function_context) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(function_context != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AST translator function context"));

    REQUIRE_OK(kefir_ast_translate_local_scope(mem, &function_context->local_context->context, function_context->module,
                                               &function_context->local_scope_layout));
    return KEFIR_OK;
}
