/*
    SPDX-License-Identifier: GPL-3.0

    Copyright (C) 2020-2025  Jevgenijs Protopopovs

    This file is part of Kefir project.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include "kefir/ast-translator/scope/translator.h"
#include "kefir/core/util.h"
#include "kefir/core/error.h"
#include "kefir/ir/type_tree.h"
#include "kefir/ast/constant_expression.h"
#include "kefir/ast/runtime.h"
#include "kefir/ast/downcast.h"
#include "kefir/ast/initializer_traversal.h"
#include "kefir/core/source_error.h"
#include "kefir/ast/type_conv.h"

static kefir_size_t resolve_identifier_offset(const struct kefir_ast_type_layout *layout) {
    if (layout->parent != NULL) {
        return resolve_identifier_offset(layout->parent) + layout->properties.relative_offset;
    } else {
        return layout->properties.relative_offset;
    }
}

static kefir_result_t local_static_identifier(struct kefir_mem *mem, struct kefir_ir_module *module,
                                              const char *function_name, const char *variable_identifier,
                                              kefir_id_t unique_id, const char **identifier) {
    int buflen = snprintf(NULL, 0, KEFIR_AST_TRANSLATOR_FUNCTION_STATIC_VARIABLE_IDENTIFIER, function_name,
                          variable_identifier, unique_id);
    char *buf = KEFIR_MALLOC(mem, sizeof(char) * (buflen + 2));
    REQUIRE(buf != NULL,
            KEFIR_SET_ERROR(KEFIR_MEMALLOC_FAILURE, "Failed to allocate local static variable identifier"));
    snprintf(buf, buflen + 1, KEFIR_AST_TRANSLATOR_FUNCTION_STATIC_VARIABLE_IDENTIFIER, function_name,
             variable_identifier, unique_id);
    *identifier = kefir_ir_module_symbol(mem, module, buf, NULL);
    KEFIR_FREE(mem, buf);
    REQUIRE(identifier != NULL,
            KEFIR_SET_ERROR(KEFIR_OBJALLOC_FAILURE, "Failed to allocate local static variable identifier"));
    return KEFIR_OK;
}

static kefir_result_t translate_pointer_to_identifier(struct kefir_mem *mem,
                                                      struct kefir_ast_constant_expression_value *value,
                                                      struct kefir_ir_module *module, struct kefir_ir_data *data,
                                                      kefir_size_t base_slot) {
    if (value->pointer.scoped_id->klass == KEFIR_AST_SCOPE_IDENTIFIER_OBJECT) {
        switch (value->pointer.scoped_id->object.storage) {
            case KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_EXTERN: {
                const char *literal = kefir_ir_module_symbol(mem, module, value->pointer.base.literal, NULL);
                REQUIRE(literal != NULL, KEFIR_SET_ERROR(KEFIR_MEMALLOC_FAILURE, "Failed to allocate IR symbol"));
                REQUIRE_OK(kefir_ir_data_set_pointer(mem, data, base_slot, literal, value->pointer.offset));
            } break;

            case KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_STATIC: {
                ASSIGN_DECL_CAST(struct kefir_ast_translator_scoped_identifier_object *, identifier_data,
                                 value->pointer.scoped_id->payload.ptr);
                const char *identifier = value->pointer.base.literal;
                if (value->pointer.scoped_id->object.defining_function != NULL) {
                    REQUIRE_OK(local_static_identifier(mem, module, value->pointer.scoped_id->object.defining_function,
                                                       identifier, identifier_data->identifier, &identifier));
                } else {
                    identifier = kefir_ir_module_symbol(mem, module, identifier, NULL);
                    REQUIRE(identifier != NULL, KEFIR_SET_ERROR(KEFIR_OBJALLOC_FAILURE,
                                                                "Failed to allocate local static variable identifier"));
                }
                REQUIRE_OK(kefir_ir_data_set_pointer(
                    mem, data, base_slot, identifier,
                    resolve_identifier_offset(identifier_data->layout) + value->pointer.offset));
            } break;

            case KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_THREAD_LOCAL:
            case KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_EXTERN_THREAD_LOCAL:
            case KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_STATIC_THREAD_LOCAL:
                return KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Unexpected non-constant initializer element");

            case KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_TYPEDEF:
            case KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_AUTO:
            case KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_REGISTER:
            case KEFIR_AST_SCOPE_IDENTIFIER_STORAGE_UNKNOWN:
                return KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Unexpected storage class of addressed variable");
        }
    } else if (value->pointer.scoped_id->klass == KEFIR_AST_SCOPE_IDENTIFIER_LABEL) {
        const char *literal = kefir_ir_module_symbol(mem, module, value->pointer.base.literal, NULL);
        REQUIRE(literal != NULL, KEFIR_SET_ERROR(KEFIR_MEMALLOC_FAILURE, "Failed to allocate IR symbol"));
        REQUIRE_OK(kefir_ir_data_set_pointer(mem, data, base_slot, literal, value->pointer.offset));
    } else {
        REQUIRE(value->pointer.scoped_id->klass == KEFIR_AST_SCOPE_IDENTIFIER_FUNCTION,
                KEFIR_SET_SOURCE_ERROR(
                    KEFIR_ANALYSIS_ERROR, NULL,
                    "Global variables can be initialized by pointer either to an object, a label or a function"));
        const char *literal = kefir_ir_module_symbol(mem, module, value->pointer.base.literal, NULL);
        REQUIRE(literal != NULL, KEFIR_SET_ERROR(KEFIR_MEMALLOC_FAILURE, "Failed to allocate IR symbol"));
        REQUIRE_OK(kefir_ir_data_set_pointer(mem, data, base_slot, literal, value->pointer.offset));
    }
    return KEFIR_OK;
}

struct designator_resolve_param {
    struct kefir_ir_type_tree *ir_type_tree;
    kefir_size_t *slot;
};

static kefir_result_t add_designated_slot(struct kefir_ast_type_layout *layout,
                                          const struct kefir_ast_designator *designator, void *payload) {
    REQUIRE(layout != NULL, KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected valid AST type layout"));
    REQUIRE(designator != NULL, KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected valid designator"));
    REQUIRE(payload != NULL, KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected valid payload"));

    ASSIGN_DECL_CAST(struct designator_resolve_param *, param, payload);

    const struct kefir_ir_type_tree_node *tree_node;
    REQUIRE_OK(kefir_ir_type_tree_at(param->ir_type_tree, layout->value, &tree_node));

    if (designator->type == KEFIR_AST_DESIGNATOR_MEMBER) {
        *param->slot += tree_node->relative_slot;
    } else if (designator->type == KEFIR_AST_DESIGNATOR_SUBSCRIPT) {
        *param->slot += tree_node->relative_slot + designator->index * tree_node->slot_width;
    } else {
        return KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Unexpected designator type");
    }
    return KEFIR_OK;
}

static kefir_result_t resolve_designated_slot(struct kefir_ast_type_layout *root,
                                              const struct kefir_ast_designator *designator,
                                              struct kefir_ir_type_tree *ir_type_tree, kefir_size_t base_slot,
                                              struct kefir_ast_type_layout **resolved_layout, kefir_size_t *slot,
                                              const struct kefir_source_location *source_location) {
    *resolved_layout = root;
    *slot = base_slot;
    if (designator != NULL) {
        struct designator_resolve_param param = {.ir_type_tree = ir_type_tree, .slot = slot};
        REQUIRE_OK(kefir_ast_type_layout_resolve(root, designator, resolved_layout, add_designated_slot, &param));
    }

    for (const struct kefir_ast_type_layout *type_layout = *resolved_layout; type_layout != NULL;
         type_layout = type_layout->parent) {

        REQUIRE(type_layout->type->tag != KEFIR_AST_TYPE_ARRAY ||
                    type_layout->type->array_type.boundary != KEFIR_AST_ARRAY_UNBOUNDED,
                KEFIR_SET_SOURCE_ERROR(KEFIR_ANALYSIS_ERROR, source_location,
                                       "Initialization of flexible array members is not supported"));
    }
    return KEFIR_OK;
}

struct traversal_param {
    struct kefir_mem *mem;
    const struct kefir_ast_context *context;
    struct kefir_ir_module *module;
    struct kefir_ast_type_layout *type_layout;
    const struct kefir_ir_type *type;
    struct kefir_ir_data *data;
    kefir_size_t base_slot;
    struct kefir_ir_type_tree ir_type_tree;
};

static const kefir_ir_string_literal_type_t StringLiteralTypes[] = {
    [KEFIR_AST_STRING_LITERAL_MULTIBYTE] = KEFIR_IR_STRING_LITERAL_MULTIBYTE,
    [KEFIR_AST_STRING_LITERAL_UNICODE8] = KEFIR_IR_STRING_LITERAL_MULTIBYTE,
    [KEFIR_AST_STRING_LITERAL_UNICODE16] = KEFIR_IR_STRING_LITERAL_UNICODE16,
    [KEFIR_AST_STRING_LITERAL_UNICODE32] = KEFIR_IR_STRING_LITERAL_UNICODE32,
    [KEFIR_AST_STRING_LITERAL_WIDE] = KEFIR_IR_STRING_LITERAL_UNICODE32};

static kefir_result_t visit_value(const struct kefir_ast_designator *designator, struct kefir_ast_node_base *expression,
                                  void *payload) {
    REQUIRE(expression != NULL, KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected valid AST expression node"));
    REQUIRE(payload != NULL, KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected valid payload"));
    ASSIGN_DECL_CAST(struct traversal_param *, param, payload);

    struct kefir_ast_type_layout *resolved_layout = NULL;
    kefir_size_t slot = 0;
    REQUIRE_OK(resolve_designated_slot(param->type_layout, designator, &param->ir_type_tree, param->base_slot,
                                       &resolved_layout, &slot, &expression->source_location));

    struct kefir_ast_compound_literal *compound_literal;
    kefir_result_t res = kefir_ast_downcast_compound_literal(expression, &compound_literal, false);
    if (res != KEFIR_NO_MATCH) {
        REQUIRE_OK(res);
        const struct kefir_ast_type *expr_type = kefir_ast_unqualified_type(expression->properties.type);
        if (!KEFIR_AST_TYPE_IS_SCALAR_TYPE(expr_type) && expr_type->tag != KEFIR_AST_TYPE_ARRAY) {
            REQUIRE(KEFIR_AST_TYPE_COMPATIBLE(param->context->type_traits, expr_type, resolved_layout->type),
                    KEFIR_SET_SOURCE_ERROR(KEFIR_ANALYSIS_ERROR, &expression->source_location,
                                           "Compound literal type mismatch"));
            REQUIRE_OK(kefir_ast_translate_data_initializer(param->mem, param->context, param->module, resolved_layout,
                                                            param->type, compound_literal->initializer, param->data,
                                                            slot));
            return KEFIR_OK;
        }
    }

    struct kefir_ast_constant_expression_value uncasted_value, value;
    REQUIRE_OK(kefir_ast_constant_expression_value_evaluate(param->mem, param->context, expression, &uncasted_value));
    REQUIRE_OK(kefir_ast_constant_expression_value_cast(param->mem, param->context, &value, &uncasted_value, expression,
                                                        resolved_layout->type));
    struct kefir_ir_typeentry *target_typeentry = kefir_ir_type_at(param->type, resolved_layout->value);
    REQUIRE(target_typeentry != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Cannot obtain target IR type entry"));
    switch (value.klass) {
        case KEFIR_AST_CONSTANT_EXPRESSION_CLASS_NONE:
            return KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Unexpected constant expression value type");

        case KEFIR_AST_CONSTANT_EXPRESSION_CLASS_INTEGER:
            switch (target_typeentry->typecode) {
                case KEFIR_IR_TYPE_INT8:
                case KEFIR_IR_TYPE_INT16:
                case KEFIR_IR_TYPE_INT32:
                case KEFIR_IR_TYPE_INT64:
                case KEFIR_IR_TYPE_BOOL:
                case KEFIR_IR_TYPE_CHAR:
                case KEFIR_IR_TYPE_SHORT:
                case KEFIR_IR_TYPE_INT:
                case KEFIR_IR_TYPE_LONG:
                case KEFIR_IR_TYPE_WORD:
                case KEFIR_IR_TYPE_BITS:
                    if (resolved_layout->bitfield) {
                        REQUIRE_OK(kefir_ir_data_set_bitfield(param->mem, param->data, slot, value.integer,
                                                              resolved_layout->bitfield_props.offset,
                                                              resolved_layout->bitfield_props.width));
                    } else {
                        REQUIRE_OK(kefir_ir_data_set_integer(param->mem, param->data, slot, value.integer));
                    }
                    break;

                case KEFIR_IR_TYPE_FLOAT32:
                    REQUIRE_OK(
                        kefir_ir_data_set_float32(param->mem, param->data, slot, (kefir_float32_t) value.integer));
                    break;

                case KEFIR_IR_TYPE_FLOAT64:
                    REQUIRE_OK(
                        kefir_ir_data_set_float64(param->mem, param->data, slot, (kefir_float64_t) value.integer));
                    break;

                case KEFIR_IR_TYPE_LONG_DOUBLE:
                    REQUIRE_OK(kefir_ir_data_set_long_double(param->mem, param->data, slot,
                                                             (kefir_long_double_t) value.integer));
                    break;

                default:
                    return KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Unexpected target IR type entry code");
            }
            break;

        case KEFIR_AST_CONSTANT_EXPRESSION_CLASS_FLOAT:
            switch (target_typeentry->typecode) {
                case KEFIR_IR_TYPE_INT8:
                case KEFIR_IR_TYPE_INT16:
                case KEFIR_IR_TYPE_INT32:
                case KEFIR_IR_TYPE_INT64:
                case KEFIR_IR_TYPE_BOOL:
                case KEFIR_IR_TYPE_CHAR:
                case KEFIR_IR_TYPE_SHORT:
                case KEFIR_IR_TYPE_INT:
                case KEFIR_IR_TYPE_LONG:
                case KEFIR_IR_TYPE_WORD:
                    REQUIRE_OK(
                        kefir_ir_data_set_integer(param->mem, param->data, slot, (kefir_int64_t) value.floating_point));
                    break;

                case KEFIR_IR_TYPE_FLOAT32:
                    REQUIRE_OK(kefir_ir_data_set_float32(param->mem, param->data, slot, value.floating_point));
                    break;

                case KEFIR_IR_TYPE_FLOAT64:
                    REQUIRE_OK(kefir_ir_data_set_float64(param->mem, param->data, slot, value.floating_point));
                    break;

                case KEFIR_IR_TYPE_LONG_DOUBLE:
                    REQUIRE_OK(kefir_ir_data_set_long_double(param->mem, param->data, slot, value.floating_point));
                    break;

                case KEFIR_IR_TYPE_COMPLEX_FLOAT32:
                    REQUIRE_OK(
                        kefir_ir_data_set_complex_float32(param->mem, param->data, slot, value.floating_point, 0.0f));
                    break;

                case KEFIR_IR_TYPE_COMPLEX_FLOAT64:
                    REQUIRE_OK(
                        kefir_ir_data_set_complex_float64(param->mem, param->data, slot, value.floating_point, 0.0));
                    break;

                case KEFIR_IR_TYPE_COMPLEX_LONG_DOUBLE:
                    REQUIRE_OK(kefir_ir_data_set_complex_long_double(param->mem, param->data, slot,
                                                                     value.floating_point, 0.0L));
                    break;

                default:
                    return KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Unexpected target IR type entry code");
            }
            break;

        case KEFIR_AST_CONSTANT_EXPRESSION_CLASS_COMPLEX_FLOAT:
            switch (target_typeentry->typecode) {
                case KEFIR_IR_TYPE_INT8:
                case KEFIR_IR_TYPE_INT16:
                case KEFIR_IR_TYPE_INT32:
                case KEFIR_IR_TYPE_INT64:
                case KEFIR_IR_TYPE_BOOL:
                case KEFIR_IR_TYPE_CHAR:
                case KEFIR_IR_TYPE_SHORT:
                case KEFIR_IR_TYPE_INT:
                case KEFIR_IR_TYPE_LONG:
                case KEFIR_IR_TYPE_WORD:
                    REQUIRE_OK(kefir_ir_data_set_integer(param->mem, param->data, slot,
                                                         (kefir_int64_t) value.complex_floating_point.real));
                    break;

                case KEFIR_IR_TYPE_FLOAT32:
                    REQUIRE_OK(
                        kefir_ir_data_set_float32(param->mem, param->data, slot, value.complex_floating_point.real));
                    break;

                case KEFIR_IR_TYPE_FLOAT64:
                    REQUIRE_OK(
                        kefir_ir_data_set_float64(param->mem, param->data, slot, value.complex_floating_point.real));
                    break;

                case KEFIR_IR_TYPE_LONG_DOUBLE:
                    REQUIRE_OK(kefir_ir_data_set_long_double(param->mem, param->data, slot,
                                                             value.complex_floating_point.real));
                    break;

                case KEFIR_IR_TYPE_COMPLEX_FLOAT32:
                    REQUIRE_OK(kefir_ir_data_set_complex_float32(param->mem, param->data, slot,
                                                                 value.complex_floating_point.real,
                                                                 value.complex_floating_point.imaginary));
                    break;

                case KEFIR_IR_TYPE_COMPLEX_FLOAT64:
                    REQUIRE_OK(kefir_ir_data_set_complex_float64(param->mem, param->data, slot,
                                                                 value.complex_floating_point.real,
                                                                 value.complex_floating_point.imaginary));
                    break;

                case KEFIR_IR_TYPE_COMPLEX_LONG_DOUBLE:
                    REQUIRE_OK(kefir_ir_data_set_complex_long_double(param->mem, param->data, slot,
                                                                     value.complex_floating_point.real,
                                                                     value.complex_floating_point.imaginary));
                    break;

                default:
                    return KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Unexpected target IR type entry code");
            }
            break;

        case KEFIR_AST_CONSTANT_EXPRESSION_CLASS_ADDRESS:
            switch (value.pointer.type) {
                case KEFIR_AST_CONSTANT_EXPRESSION_POINTER_IDENTIFER:
                    REQUIRE_OK(translate_pointer_to_identifier(param->mem, &value, param->module, param->data, slot));
                    break;

                case KEFIR_AST_CONSTANT_EXPRESSION_POINTER_INTEGER:
                    REQUIRE_OK(kefir_ir_data_set_integer(param->mem, param->data, slot,
                                                         value.pointer.base.integral + value.pointer.offset));
                    break;

                case KEFIR_AST_CONSTANT_EXPRESSION_POINTER_LITERAL: {
                    kefir_id_t id;
                    REQUIRE_OK(kefir_ir_module_string_literal(
                        param->mem, param->module, StringLiteralTypes[value.pointer.base.string.type], true,
                        value.pointer.base.string.content, value.pointer.base.string.length, &id));
                    REQUIRE_OK(
                        kefir_ir_data_set_string_pointer(param->mem, param->data, slot, id, value.pointer.offset));
                } break;
            }
            break;
    }

    return KEFIR_OK;
}

static kefir_result_t visit_string_literal(const struct kefir_ast_designator *designator,
                                           struct kefir_ast_node_base *expression, kefir_ast_string_literal_type_t type,
                                           const void *string, kefir_size_t length, void *payload) {
    UNUSED(expression);
    REQUIRE(string != NULL, KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected valid string literal"));
    REQUIRE(payload != NULL, KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected valid payload"));
    ASSIGN_DECL_CAST(struct traversal_param *, param, payload);

    kefir_id_t string_id;
    REQUIRE_OK(kefir_ir_module_string_literal(param->mem, param->module, StringLiteralTypes[type], false, string,
                                              length, &string_id));

    kefir_ir_string_literal_type_t ir_string_type;
    kefir_bool_t public;
    const void *string_content = NULL;
    kefir_size_t string_length = 0;
    REQUIRE_OK(kefir_ir_module_get_string_literal(param->module, string_id, &ir_string_type, &public, &string_content,
                                                  &string_length));

    struct kefir_ast_type_layout *resolved_layout = NULL;
    kefir_size_t slot = 0;
    REQUIRE_OK(resolve_designated_slot(param->type_layout, designator, &param->ir_type_tree, param->base_slot,
                                       &resolved_layout, &slot, &expression->source_location));

    if (resolved_layout->type->tag == KEFIR_AST_TYPE_ARRAY &&
        (resolved_layout->type->array_type.boundary == KEFIR_AST_ARRAY_BOUNDED ||
         resolved_layout->type->array_type.boundary == KEFIR_AST_ARRAY_BOUNDED_STATIC)) {
        string_length =
            MIN(string_length, (kefir_size_t) resolved_layout->type->array_type.const_length->value.integer);
    }

    REQUIRE_OK(kefir_ir_data_set_string(param->mem, param->data, slot, StringLiteralTypes[type], string_content,
                                        string_length));
    return KEFIR_OK;
}

static kefir_result_t visit_initializer_list(const struct kefir_ast_designator *designator,
                                             const struct kefir_ast_initializer *initializer, void *payload) {
    REQUIRE(initializer != NULL, KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected valid AST initializer"));
    REQUIRE(payload != NULL, KEFIR_SET_ERROR(KEFIR_INTERNAL_ERROR, "Expected valid payload"));
    ASSIGN_DECL_CAST(struct traversal_param *, param, payload);

    struct kefir_ast_type_layout *resolved_layout = NULL;
    kefir_size_t slot = 0;
    REQUIRE_OK(resolve_designated_slot(param->type_layout, designator, &param->ir_type_tree, param->base_slot,
                                       &resolved_layout, &slot, &initializer->source_location));

    REQUIRE_OK(kefir_ast_translate_data_initializer(param->mem, param->context, param->module, resolved_layout,
                                                    param->type, initializer, param->data, slot));
    return KEFIR_OK;
}

kefir_result_t kefir_ast_translate_data_initializer(struct kefir_mem *mem, const struct kefir_ast_context *context,
                                                    struct kefir_ir_module *module,
                                                    struct kefir_ast_type_layout *type_layout,
                                                    const struct kefir_ir_type *type,
                                                    const struct kefir_ast_initializer *initializer,
                                                    struct kefir_ir_data *data, kefir_size_t base_slot) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(context != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AST context"));
    REQUIRE(module != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid IR module"));
    REQUIRE(type_layout != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AST type layout"));
    REQUIRE(type != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid IR type"));
    REQUIRE(initializer != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AST initializer"));
    REQUIRE(data != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid IR data"));

    struct traversal_param param = {.mem = mem,
                                    .context = context,
                                    .module = module,
                                    .type_layout = type_layout,
                                    .type = type,
                                    .data = data,
                                    .base_slot = base_slot};

    struct kefir_ast_initializer_traversal initializer_traversal;
    KEFIR_AST_INITIALIZER_TRAVERSAL_INIT(&initializer_traversal);
    initializer_traversal.visit_value = visit_value;
    initializer_traversal.visit_string_literal = visit_string_literal;
    initializer_traversal.visit_initializer_list = visit_initializer_list;
    initializer_traversal.payload = &param;

    REQUIRE_OK(kefir_ir_type_tree_init(mem, type, &param.ir_type_tree));
    kefir_result_t res =
        kefi_ast_traverse_initializer(mem, context, initializer, type_layout->type, &initializer_traversal);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_ir_type_tree_free(mem, &param.ir_type_tree);
        return res;
    });
    REQUIRE_OK(kefir_ir_type_tree_free(mem, &param.ir_type_tree));
    return KEFIR_OK;
}
