/*
    SPDX-License-Identifier: GPL-3.0

    Copyright (C) 2020-2025  Jevgenijs Protopopovs

    This file is part of Kefir project.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "kefir/preprocessor/user_macro.h"
#include "kefir/preprocessor/preprocessor.h"
#include "kefir/preprocessor/util.h"
#include "kefir/core/util.h"
#include "kefir/core/error.h"
#include "kefir/core/hashtree.h"
#include "kefir/core/source_error.h"
#include "kefir/preprocessor/format.h"
#include "kefir/preprocessor/token_sequence.h"
#include "kefir/lexer/format.h"
#include <string.h>

static kefir_result_t user_macro_argc(const struct kefir_preprocessor_macro *macro, kefir_size_t *argc_ptr,
                                      kefir_bool_t *vararg_ptr) {
    REQUIRE(macro != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor macro"));
    REQUIRE(argc_ptr != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to argument count"));
    REQUIRE(vararg_ptr != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to vararg flag"));
    ASSIGN_DECL_CAST(struct kefir_preprocessor_user_macro *, user_macro, macro->payload);

    REQUIRE(macro->type == KEFIR_PREPROCESSOR_MACRO_FUNCTION,
            KEFIR_SET_ERROR(KEFIR_INVALID_REQUEST, "Unable to retrieve argument count of object macro"));
    *argc_ptr = kefir_list_length(&user_macro->parameters);
    *vararg_ptr = user_macro->vararg;
    return KEFIR_OK;
}

static const char *user_macro_vararg_name(const struct kefir_preprocessor_user_macro *user_macro) {
    return user_macro->vararg_parameter != NULL ? user_macro->vararg_parameter : "__VA_ARGS__";
}

static kefir_result_t build_function_macro_args_tree(struct kefir_mem *mem,
                                                     const struct kefir_preprocessor_user_macro *user_macro,
                                                     const struct kefir_list *args, struct kefir_hashtree *arg_tree) {
    const struct kefir_list_entry *id_iter = kefir_list_head(&user_macro->parameters);
    const struct kefir_list_entry *arg_iter = kefir_list_head(args);
    for (; id_iter != NULL; kefir_list_next(&id_iter), kefir_list_next(&arg_iter)) {
        REQUIRE(arg_iter != NULL,
                KEFIR_SET_ERROR(KEFIR_OUT_OF_BOUNDS, "Provided macro arguments do no match macro parameters"));
        const char *arg_identifier = id_iter->value;
        struct kefir_token_buffer *arg_value = arg_iter->value;

        REQUIRE_OK(kefir_hashtree_insert(mem, arg_tree, (kefir_hashtree_key_t) arg_identifier,
                                         (kefir_hashtree_value_t) arg_value));
    }

    if (user_macro->vararg) {
        REQUIRE(arg_iter != NULL,
                KEFIR_SET_ERROR(KEFIR_OUT_OF_BOUNDS, "Provided macro arguments do no match macro parameters"));
        const struct kefir_token_buffer *arg_value = arg_iter->value;
        const char *vararg_parameter = user_macro_vararg_name(user_macro);
        REQUIRE_OK(kefir_hashtree_insert(mem, arg_tree, (kefir_hashtree_key_t) vararg_parameter,
                                         (kefir_hashtree_value_t) arg_value));
    }
    return KEFIR_OK;
}

static kefir_result_t fn_macro_parameter_substitution(struct kefir_mem *mem, struct kefir_preprocessor *preprocessor,
                                                      struct kefir_token_allocator *token_allocator,
                                                      const struct kefir_hashtree *arg_tree,
                                                      struct kefir_token_buffer *buffer, const char *identifier) {
    struct kefir_hashtree_node *node;
    kefir_result_t res = kefir_hashtree_at(arg_tree, (kefir_hashtree_key_t) identifier, &node);
    if (res != KEFIR_NOT_FOUND) {
        REQUIRE_OK(res);

        ASSIGN_DECL_CAST(const struct kefir_token_buffer *, arg_buffer, node->value);
        struct kefir_token_buffer subst_buffer;
        REQUIRE_OK(kefir_token_buffer_init(&subst_buffer));
        kefir_result_t res = kefir_token_buffer_copy(mem, &subst_buffer, arg_buffer);
        REQUIRE_CHAIN(&res, kefir_preprocessor_run_substitutions(mem, preprocessor, token_allocator, &subst_buffer,
                                                                 NULL, KEFIR_PREPROCESSOR_SUBSTITUTION_NORMAL));
        REQUIRE_CHAIN(&res, kefir_token_buffer_copy(mem, buffer, &subst_buffer));
        REQUIRE_ELSE(res == KEFIR_OK, {
            kefir_token_buffer_free(mem, &subst_buffer);
            return res;
        });
        REQUIRE_OK(kefir_token_buffer_free(mem, &subst_buffer));
    } else {
        return KEFIR_SET_ERROR(KEFIR_NO_MATCH, "Unable to substitute function macro parameter");
    }
    return KEFIR_OK;
}

static kefir_result_t fn_macro_parameter_stringification(struct kefir_mem *mem, const struct kefir_hashtree *arg_tree,
                                                         struct kefir_token_allocator *token_allocator,
                                                         struct kefir_token_buffer *buffer, const char *identifier,
                                                         const struct kefir_source_location *source_location) {
    struct kefir_hashtree_node *node;
    kefir_result_t res = kefir_hashtree_at(arg_tree, (kefir_hashtree_key_t) identifier, &node);
    if (res == KEFIR_NOT_FOUND) {
        return KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, source_location,
                                      "Expected # to be followed by macro parameter");
    } else {
        REQUIRE_OK(res);
    }
    ASSIGN_DECL_CAST(const struct kefir_token_buffer *, arg_buffer, node->value);

    char *string = NULL;
    kefir_size_t string_sz = 0;
    REQUIRE_OK(kefir_preprocessor_format_string(mem, &string, &string_sz, arg_buffer,
                                                KEFIR_PREPROCESSOR_WHITESPACE_FORMAT_SINGLE_SPACE));

    struct kefir_token *allocated_token;
    res = kefir_token_allocator_allocate_empty(mem, token_allocator, &allocated_token);
    REQUIRE_CHAIN(&res, kefir_token_new_string_literal_raw_from_escaped_multibyte(
                            mem, KEFIR_STRING_LITERAL_TOKEN_MULTIBYTE, string, string_sz - 1, allocated_token));
    REQUIRE_ELSE(res == KEFIR_OK, {
        KEFIR_FREE(mem, string);
        return res;
    });
    KEFIR_FREE(mem, string);
    allocated_token->source_location = *source_location;
    REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, allocated_token));
    return KEFIR_OK;
}

static const struct kefir_token PlacemakerToken = {.klass = KEFIR_TOKEN_PP_PLACEMAKER};

static kefir_result_t concat_resolve_left_argument(struct kefir_mem *mem, const struct kefir_hashtree *arg_tree,
                                                   struct kefir_token_buffer *buffer, const struct kefir_token *arg,
                                                   const struct kefir_token **result) {
    if (arg->klass == KEFIR_TOKEN_IDENTIFIER && arg_tree != NULL) {
        struct kefir_hashtree_node *node;
        kefir_result_t res = kefir_hashtree_at(arg_tree, (kefir_hashtree_key_t) arg->identifier, &node);
        if (res != KEFIR_NOT_FOUND) {
            REQUIRE_OK(res);
            ASSIGN_DECL_CAST(const struct kefir_token_buffer *, arg_buffer, node->value);

            const kefir_size_t arg_buffer_length = kefir_token_buffer_length(arg_buffer);
            if (arg_buffer_length == 0) {
                *result = &PlacemakerToken;
            } else {
                for (kefir_size_t i = 0; i < arg_buffer_length - 1; i++) {
                    REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, kefir_token_buffer_at(arg_buffer, i)));
                }
                *result = kefir_token_buffer_at(arg_buffer, arg_buffer_length - 1);
            }
            return KEFIR_OK;
        }
    }

    *result = arg;
    return KEFIR_OK;
}

static kefir_result_t concat_resolve_right_argument(struct kefir_mem *mem, const struct kefir_hashtree *arg_tree,
                                                    struct kefir_token_buffer *buffer, const struct kefir_token *arg,
                                                    const struct kefir_token **result) {
    if (arg->klass == KEFIR_TOKEN_IDENTIFIER && arg_tree != NULL) {
        struct kefir_hashtree_node *node;
        kefir_result_t res = kefir_hashtree_at(arg_tree, (kefir_hashtree_key_t) arg->identifier, &node);
        if (res != KEFIR_NOT_FOUND) {
            REQUIRE_OK(res);
            ASSIGN_DECL_CAST(const struct kefir_token_buffer *, arg_buffer, node->value);

            const kefir_size_t arg_buffer_length = kefir_token_buffer_length(arg_buffer);
            if (arg_buffer_length == 0) {
                *result = &PlacemakerToken;
            } else {
                for (kefir_size_t i = 1; i < arg_buffer_length; i++) {
                    REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, kefir_token_buffer_at(arg_buffer, i)));
                }
                *result = kefir_token_buffer_at(arg_buffer, 0);
            }
            return KEFIR_OK;
        }
    }

    *result = arg;
    return KEFIR_OK;
}

static char *concat_strings(struct kefir_mem *mem, const char *str1, const char *str2) {

    kefir_size_t sz = strlen(str1) + strlen(str2) + 1;
    char *result = KEFIR_MALLOC(mem, sz);
    REQUIRE(result != NULL, NULL);
    strcpy(result, str1);
    strcat(result, str2);
    return result;
}

static kefir_result_t concat_tokens_impl(struct kefir_mem *mem, struct kefir_string_pool *symbols,
                                         struct kefir_token_allocator *token_allocator, const struct kefir_token *left,
                                         const struct kefir_token *right, struct kefir_token_buffer *buffer) {
    REQUIRE(
        right->klass == KEFIR_TOKEN_IDENTIFIER || right->klass == KEFIR_TOKEN_PP_NUMBER ||
            (right->klass == KEFIR_TOKEN_CONSTANT && right->constant.type == KEFIR_CONSTANT_TOKEN_CHAR) ||
            (right->klass == KEFIR_TOKEN_STRING_LITERAL &&
             right->string_literal.type == KEFIR_STRING_LITERAL_TOKEN_MULTIBYTE && right->string_literal.raw_literal),
        KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &right->source_location,
                               "Expected either identifier, character or preprocessor number"));
    kefir_result_t res;

    if (right->klass == KEFIR_TOKEN_CONSTANT && right->constant.type == KEFIR_CONSTANT_TOKEN_CHAR) {
        REQUIRE(left->klass == KEFIR_TOKEN_IDENTIFIER,
                KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location, "Expected character literal prefix"));

        struct kefir_token *allocated_token;
        REQUIRE_OK(kefir_token_allocator_allocate_empty(mem, token_allocator, &allocated_token));

        if (strcmp(left->identifier, "L") == 0) {
            REQUIRE_OK(kefir_token_new_constant_wide_char(right->constant.character, allocated_token));
        } else if (strcmp(left->identifier, "u") == 0) {
            REQUIRE_OK(
                kefir_token_new_constant_unicode16_char((kefir_char16_t) right->constant.character, allocated_token));
        } else if (strcmp(left->identifier, "U") == 0) {
            REQUIRE_OK(
                kefir_token_new_constant_unicode32_char((kefir_char32_t) right->constant.character, allocated_token));
        } else {
            return KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location,
                                          "Expected character literal prefix");
        }

        REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, allocated_token));
        return KEFIR_OK;
    }

    if (right->klass == KEFIR_TOKEN_STRING_LITERAL &&
        right->string_literal.type == KEFIR_STRING_LITERAL_TOKEN_MULTIBYTE && right->string_literal.raw_literal) {
        REQUIRE(left->klass == KEFIR_TOKEN_IDENTIFIER,
                KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location, "Expected string literal prefix"));

        struct kefir_token *allocated_token;
        REQUIRE_OK(kefir_token_allocator_allocate_empty(mem, token_allocator, &allocated_token));

        if (strcmp(left->identifier, "L") == 0) {
            REQUIRE_OK(kefir_token_new_string_literal_raw(mem, KEFIR_STRING_LITERAL_TOKEN_WIDE,
                                                          right->string_literal.literal, right->string_literal.length,
                                                          allocated_token));
        } else if (strcmp(left->identifier, "u8") == 0) {
            REQUIRE_OK(kefir_token_new_string_literal_raw(mem, KEFIR_STRING_LITERAL_TOKEN_UNICODE8,
                                                          right->string_literal.literal, right->string_literal.length,
                                                          allocated_token));
        } else if (strcmp(left->identifier, "u") == 0) {
            REQUIRE_OK(kefir_token_new_string_literal_raw(mem, KEFIR_STRING_LITERAL_TOKEN_UNICODE16,
                                                          right->string_literal.literal, right->string_literal.length,
                                                          allocated_token));
        } else if (strcmp(left->identifier, "U") == 0) {
            REQUIRE_OK(kefir_token_new_string_literal_raw(mem, KEFIR_STRING_LITERAL_TOKEN_UNICODE32,
                                                          right->string_literal.literal, right->string_literal.length,
                                                          allocated_token));
        } else {
            return KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location, "Expected string literal prefix");
        }

        REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, allocated_token));
        return KEFIR_OK;
    }

    const char *left_str = left->klass == KEFIR_TOKEN_IDENTIFIER ? left->identifier : left->pp_number.number_literal;
    const char *right_str =
        right->klass == KEFIR_TOKEN_IDENTIFIER ? right->identifier : right->pp_number.number_literal;
    char *result = concat_strings(mem, left_str, right_str);
    REQUIRE(result != NULL, KEFIR_SET_ERROR(KEFIR_MEMALLOC_FAILURE, "Failed to allocate identifier string"));

    struct kefir_token *allocated_token;
    res = kefir_token_allocator_allocate_empty(mem, token_allocator, &allocated_token);
    if (left->klass == KEFIR_TOKEN_IDENTIFIER) {
        REQUIRE_CHAIN(&res, kefir_token_new_identifier(mem, symbols, result, allocated_token));
    } else {
        REQUIRE_CHAIN(&res, kefir_token_new_pp_number(mem, result, strlen(result), allocated_token));
    }
    REQUIRE_ELSE(res == KEFIR_OK, {
        KEFIR_FREE(mem, result);
        return res;
    });
    allocated_token->source_location = left->source_location;
    KEFIR_FREE(mem, result);

    REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, allocated_token));
    return KEFIR_OK;
}

static kefir_result_t macro_concat_punctuators(struct kefir_mem *mem, struct kefir_token_allocator *token_allocator,
                                               struct kefir_token_buffer *buffer, kefir_punctuator_token_t punctuator1,
                                               kefir_punctuator_token_t punctuator2) {
    static const struct {
        kefir_punctuator_token_t left;
        kefir_punctuator_token_t right;
        kefir_punctuator_token_t result;
    } Combinations[] = {
        {KEFIR_PUNCTUATOR_MINUS, KEFIR_PUNCTUATOR_GREATER_THAN, KEFIR_PUNCTUATOR_RIGHT_ARROW},
        {KEFIR_PUNCTUATOR_MINUS, KEFIR_PUNCTUATOR_MINUS, KEFIR_PUNCTUATOR_DOUBLE_MINUS},
        {KEFIR_PUNCTUATOR_PLUS, KEFIR_PUNCTUATOR_PLUS, KEFIR_PUNCTUATOR_DOUBLE_PLUS},
        {KEFIR_PUNCTUATOR_LESS_THAN, KEFIR_PUNCTUATOR_LESS_THAN, KEFIR_PUNCTUATOR_LEFT_SHIFT},
        {KEFIR_PUNCTUATOR_GREATER_THAN, KEFIR_PUNCTUATOR_GREATER_THAN, KEFIR_PUNCTUATOR_RIGHT_SHIFT},
        {KEFIR_PUNCTUATOR_LESS_THAN, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_LESS_OR_EQUAL},
        {KEFIR_PUNCTUATOR_GREATER_THAN, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_GREATER_OR_EQUAL},
        {KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_EQUAL},
        {KEFIR_PUNCTUATOR_EXCLAMATION_MARK, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_NOT_EQUAL},
        {KEFIR_PUNCTUATOR_AMPERSAND, KEFIR_PUNCTUATOR_AMPERSAND, KEFIR_PUNCTUATOR_DOUBLE_AMPERSAND},
        {KEFIR_PUNCTUATOR_VBAR, KEFIR_PUNCTUATOR_VBAR, KEFIR_PUNCTUATOR_DOUBLE_VBAR},
        {KEFIR_PUNCTUATOR_STAR, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN_MULTIPLY},
        {KEFIR_PUNCTUATOR_SLASH, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN_DIVIDE},
        {KEFIR_PUNCTUATOR_PERCENT, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN_MODULO},
        {KEFIR_PUNCTUATOR_PLUS, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN_ADD},
        {KEFIR_PUNCTUATOR_MINUS, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN_SUBTRACT},
        {KEFIR_PUNCTUATOR_LEFT_SHIFT, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN_SHIFT_LEFT},
        {KEFIR_PUNCTUATOR_LESS_THAN, KEFIR_PUNCTUATOR_LESS_OR_EQUAL, KEFIR_PUNCTUATOR_ASSIGN_SHIFT_LEFT},
        {KEFIR_PUNCTUATOR_RIGHT_SHIFT, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN_SHIFT_RIGHT},
        {KEFIR_PUNCTUATOR_GREATER_THAN, KEFIR_PUNCTUATOR_GREATER_OR_EQUAL, KEFIR_PUNCTUATOR_ASSIGN_SHIFT_RIGHT},
        {KEFIR_PUNCTUATOR_AMPERSAND, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN_AND},
        {KEFIR_PUNCTUATOR_CARET, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN_XOR},
        {KEFIR_PUNCTUATOR_VBAR, KEFIR_PUNCTUATOR_ASSIGN, KEFIR_PUNCTUATOR_ASSIGN_OR},
        {KEFIR_PUNCTUATOR_HASH, KEFIR_PUNCTUATOR_HASH, KEFIR_PUNCTUATOR_DOUBLE_HASH},
        {KEFIR_PUNCTUATOR_LESS_THAN, KEFIR_PUNCTUATOR_COLON, KEFIR_PUNCTUATOR_DIGRAPH_LEFT_BRACKET},
        {KEFIR_PUNCTUATOR_COLON, KEFIR_PUNCTUATOR_GREATER_THAN, KEFIR_PUNCTUATOR_DIGRAPH_RIGHT_BRACKET},
        {KEFIR_PUNCTUATOR_LESS_THAN, KEFIR_PUNCTUATOR_PERCENT, KEFIR_PUNCTUATOR_DIGRAPH_LEFT_BRACE},
        {KEFIR_PUNCTUATOR_PERCENT, KEFIR_PUNCTUATOR_GREATER_THAN, KEFIR_PUNCTUATOR_DIGRAPH_RIGHT_BRACE},
        {KEFIR_PUNCTUATOR_PERCENT, KEFIR_PUNCTUATOR_COLON, KEFIR_PUNCTUATOR_DIGRAPH_HASH},
        {KEFIR_PUNCTUATOR_DIGRAPH_HASH, KEFIR_PUNCTUATOR_DIGRAPH_HASH, KEFIR_PUNCTUATOR_DIGRAPH_DOUBLE_HASH}};
    static const kefir_size_t CombinationCount = sizeof(Combinations) / sizeof(Combinations[0]);
    for (kefir_size_t i = 0; i < CombinationCount; i++) {
        if (Combinations[i].left == punctuator1 && Combinations[i].right == punctuator2) {
            struct kefir_token *allocated_token;
            REQUIRE_OK(kefir_token_allocator_allocate_empty(mem, token_allocator, &allocated_token));
            REQUIRE_OK(kefir_token_new_punctuator(Combinations[i].result, allocated_token));
            REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, allocated_token));
            return KEFIR_OK;
        }
    }
    return KEFIR_SET_ERROR(KEFIR_NO_MATCH, "Unable to concatenate punctuators");
}

static kefir_result_t concat_extension(struct kefir_mem *mem, const struct kefir_token *left,
                                       const struct kefir_token *right, struct kefir_token_allocator *token_allocator,
                                       struct kefir_token_buffer *buffer,
                                       const struct kefir_token_extension_class *klass) {
    REQUIRE(klass != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_STATE, "Invalid extension token"));
    struct kefir_token *allocated_token;
    REQUIRE_OK(kefir_token_allocator_allocate_empty(mem, token_allocator, &allocated_token));
    REQUIRE_OK(klass->concat(mem, left, right, allocated_token));
    REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, allocated_token));
    return KEFIR_OK;
}

static kefir_result_t macro_concatenation_impl(struct kefir_mem *mem, struct kefir_string_pool *symbols,
                                               struct kefir_token_allocator *token_allocator,
                                               struct kefir_token_buffer *buffer, const struct kefir_token *left,
                                               const struct kefir_token *right) {
    if (left->klass == KEFIR_TOKEN_PP_PLACEMAKER) {
        REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, right));
    } else if (right->klass == KEFIR_TOKEN_PP_PLACEMAKER) {
        REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, left));
    } else {
        switch (left->klass) {
            case KEFIR_TOKEN_SENTINEL:
                return KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location, "Unexpected sentinel token");

            case KEFIR_TOKEN_KEYWORD:
                return KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location, "Unexpected keyword token");

            case KEFIR_TOKEN_IDENTIFIER:
                if (right->klass != KEFIR_TOKEN_EXTENSION) {
                    REQUIRE_OK(concat_tokens_impl(mem, symbols, token_allocator, left, right, buffer));
                } else {
                    REQUIRE_OK(concat_extension(mem, left, right, token_allocator, buffer, right->extension.klass));
                }
                break;

            case KEFIR_TOKEN_CONSTANT:
                return KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location, "Unexpected constant token");

            case KEFIR_TOKEN_STRING_LITERAL:
                return KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location,
                                              "Unexpected string literal token");

            case KEFIR_TOKEN_PUNCTUATOR: {
                if (right->klass != KEFIR_TOKEN_EXTENSION) {
                    REQUIRE(right->klass == KEFIR_TOKEN_PUNCTUATOR,
                            KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &right->source_location,
                                                   "Expected punctuator token"));
                    kefir_result_t res =
                        macro_concat_punctuators(mem, token_allocator, buffer, left->punctuator, right->punctuator);
                    if (res == KEFIR_NO_MATCH) {
                        return KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location,
                                                      "Unable to concatenate punctuators");
                    } else {
                        REQUIRE_OK(res);
                    }
                } else {
                    REQUIRE_OK(concat_extension(mem, left, right, token_allocator, buffer, right->extension.klass));
                }
            } break;

            case KEFIR_TOKEN_PP_WHITESPACE:
                return KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location,
                                              "Unexpected preprocessor whitespace token");

            case KEFIR_TOKEN_PP_NUMBER:
                if (right->klass != KEFIR_TOKEN_EXTENSION) {
                    REQUIRE_OK(concat_tokens_impl(mem, symbols, token_allocator, left, right, buffer));
                } else {
                    REQUIRE_OK(concat_extension(mem, left, right, token_allocator, buffer, right->extension.klass));
                }
                break;

            case KEFIR_TOKEN_PP_HEADER_NAME:
                return KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &left->source_location,
                                              "Unexpected preprocessor header name token");

            case KEFIR_TOKEN_PP_PLACEMAKER:
                return KEFIR_SET_SOURCE_ERROR(KEFIR_INTERNAL_ERROR, &left->source_location,
                                              "Unexpected placemaker token");

            case KEFIR_TOKEN_EXTENSION:
                REQUIRE_OK(concat_extension(mem, left, right, token_allocator, buffer, left->extension.klass));
                break;
        }
    }
    return KEFIR_OK;
}

static kefir_result_t macro_concatenation(struct kefir_mem *mem, struct kefir_preprocessor *preprocessor,
                                          const struct kefir_hashtree *arg_tree,
                                          const struct kefir_preprocessor_user_macro *user_macro,
                                          struct kefir_string_pool *symbols,
                                          struct kefir_token_allocator *token_allocator,
                                          struct kefir_token_buffer *buffer, const struct kefir_token *arg1,
                                          const struct kefir_token *arg2, kefir_bool_t resolve_arg1) {
    if (preprocessor->context->preprocessor_config->va_args_concat && user_macro->vararg &&
        arg1->klass == KEFIR_TOKEN_PUNCTUATOR && arg1->punctuator == KEFIR_PUNCTUATOR_COMMA &&
        arg2->klass == KEFIR_TOKEN_IDENTIFIER && strcmp(arg2->identifier, user_macro_vararg_name(user_macro)) == 0) {
        // Special case
        struct kefir_hashtree_node *node = NULL;
        kefir_result_t res = kefir_hashtree_at(arg_tree, (kefir_hashtree_key_t) arg2->identifier, &node);
        REQUIRE(res == KEFIR_OK, KEFIR_SET_SOURCE_ERRORF(KEFIR_LEXER_ERROR, &arg2->source_location,
                                                         "Unable to find %s in macro parameters", arg2->identifier));
        ASSIGN_DECL_CAST(const struct kefir_token_buffer *, arg_buffer, node->value);
        const kefir_size_t arg_buffer_length = kefir_token_buffer_length(arg_buffer);
        if (arg_buffer_length == 0) {
            REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, &PlacemakerToken));
        } else {
            REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, arg1));
            REQUIRE_OK(kefir_token_buffer_copy(mem, buffer, arg_buffer));
        }
        return KEFIR_OK;
    }

    const struct kefir_token *left;
    const struct kefir_token *right;
    struct kefir_token_buffer tail;
    if (resolve_arg1) {
        REQUIRE_OK(concat_resolve_left_argument(mem, arg_tree, buffer, arg1, &left));
    } else {
        left = arg1;
    }
    REQUIRE_OK(kefir_token_buffer_init(&tail));
    kefir_result_t res = concat_resolve_right_argument(mem, arg_tree, &tail, arg2, &right);
    REQUIRE_CHAIN(&res, macro_concatenation_impl(mem, symbols, token_allocator, buffer, left, right));
    REQUIRE_CHAIN(&res, kefir_token_buffer_insert(mem, buffer, &tail));
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_token_buffer_free(mem, &tail);
        return res;
    });
    REQUIRE_OK(kefir_token_buffer_free(mem, &tail));
    return KEFIR_OK;
}

static kefir_result_t match_concatenation(
    struct kefir_mem *mem, struct kefir_preprocessor *preprocessor, struct kefir_string_pool *symbols,
    struct kefir_token_allocator *token_allocator, const struct kefir_preprocessor_user_macro *user_macro,
    struct kefir_preprocessor_token_sequence *seq, struct kefir_token_buffer *output_buffer,
    const struct kefir_hashtree *arg_tree, const struct kefir_token *left_operand, kefir_bool_t resolve_left_operand) {
    if (left_operand->klass == KEFIR_TOKEN_PP_WHITESPACE) {
        return KEFIR_SET_ERROR(KEFIR_NO_MATCH, "Unable to match preprocessor concatenation");
    }

    const struct kefir_token *next_nonws_token = NULL;
    struct kefir_token_buffer whitespaces;
    REQUIRE_OK(kefir_token_buffer_init(&whitespaces));
    kefir_result_t res = kefir_preprocessor_token_sequence_skip_whitespaces(mem, seq, &next_nonws_token, &whitespaces);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_token_buffer_free(mem, &whitespaces);
        return res;
    });
    if (next_nonws_token == NULL || next_nonws_token->klass != KEFIR_TOKEN_PUNCTUATOR ||
        next_nonws_token->punctuator != KEFIR_PUNCTUATOR_DOUBLE_HASH) {
        res = kefir_preprocessor_token_sequence_push_front(mem, seq, &whitespaces,
                                                           KEFIR_PREPROCESSOR_TOKEN_DESTINATION_NORMAL);
        REQUIRE_ELSE(res == KEFIR_OK, {
            kefir_token_buffer_free(mem, &whitespaces);
            return res;
        });
        REQUIRE_OK(kefir_token_buffer_free(mem, &whitespaces));
        return KEFIR_SET_ERROR(KEFIR_NO_MATCH, "Unable to match preprocessor concatenation");
    }
    struct kefir_source_location source_location = next_nonws_token->source_location;
    REQUIRE_OK(kefir_token_buffer_free(mem, &whitespaces));
    REQUIRE_OK(kefir_preprocessor_token_sequence_next(mem, seq, NULL, NULL));

    const struct kefir_token *right_operand = NULL;
    REQUIRE_OK(kefir_preprocessor_token_sequence_skip_whitespaces(mem, seq, &right_operand, NULL));
    REQUIRE(
        right_operand != NULL,
        KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &source_location,
                               "Preprocessor concatenation operator cannot be placed at the end of replacement list"));

    REQUIRE_OK(macro_concatenation(mem, preprocessor, arg_tree, user_macro, symbols, token_allocator, output_buffer,
                                   left_operand, right_operand, resolve_left_operand));
    REQUIRE_OK(kefir_preprocessor_token_sequence_next(mem, seq, NULL, NULL));

    for (; resolve_left_operand;) {
        left_operand = kefir_token_buffer_at(output_buffer, kefir_token_buffer_length(output_buffer) - 1);
        REQUIRE_OK(kefir_token_buffer_pop(mem, output_buffer));

        res = match_concatenation(mem, preprocessor, symbols, token_allocator, user_macro, seq, output_buffer, arg_tree,
                                  left_operand, false);
        if (res == KEFIR_NO_MATCH) {
            REQUIRE_OK(kefir_token_buffer_emplace(mem, output_buffer, left_operand));
            break;
        } else {
            REQUIRE_OK(res);
        }
    }
    return KEFIR_OK;
}

static kefir_result_t run_replacement_list_substitutions_impl(
    struct kefir_mem *mem, struct kefir_preprocessor *preprocessor, struct kefir_string_pool *symbols,
    struct kefir_token_allocator *token_allocator, const struct kefir_preprocessor_user_macro *user_macro,
    struct kefir_preprocessor_token_sequence *seq, const struct kefir_hashtree *arg_tree,
    struct kefir_token_buffer *result) {
    while (true) {
        const struct kefir_token *current_token;
        kefir_preprocessor_token_destination_t current_token_destination;
        kefir_result_t res =
            kefir_preprocessor_token_sequence_next(mem, seq, &current_token, &current_token_destination);
        REQUIRE_CHAIN_SET(
            &res, current_token_destination == KEFIR_PREPROCESSOR_TOKEN_DESTINATION_NORMAL,
            KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &current_token->source_location, "Unexpected include directive"));
        if (res == KEFIR_ITERATOR_END) {
            break;
        } else {
            REQUIRE_OK(res);
        }

        kefir_bool_t matched = false;
        res = match_concatenation(mem, preprocessor, symbols, token_allocator, user_macro, seq, result, arg_tree,
                                  current_token, true);
        if (res != KEFIR_NO_MATCH) {
            REQUIRE_OK(res);
            matched = true;
        }

        if (!matched && arg_tree != NULL && current_token->klass == KEFIR_TOKEN_IDENTIFIER) {
            res = fn_macro_parameter_substitution(mem, preprocessor, token_allocator, arg_tree, result,
                                                  current_token->identifier);
            if (res != KEFIR_NO_MATCH) {
                REQUIRE_OK(res);
                matched = true;
            }
        }

        if (!matched && arg_tree != NULL && current_token->klass == KEFIR_TOKEN_PUNCTUATOR &&
            current_token->punctuator == KEFIR_PUNCTUATOR_HASH) {
            struct kefir_source_location source_location = current_token->source_location;
            const struct kefir_token *param_token = NULL;
            REQUIRE_OK(kefir_preprocessor_token_sequence_skip_whitespaces(mem, seq, &param_token, NULL));
            REQUIRE(param_token != NULL && param_token->klass == KEFIR_TOKEN_IDENTIFIER,
                    KEFIR_SET_SOURCE_ERROR(KEFIR_LEXER_ERROR, &source_location,
                                           "Expected # to be followed by macro parameter"));
            const char *identifier = param_token->identifier;
            REQUIRE_OK(fn_macro_parameter_stringification(mem, arg_tree, token_allocator, result, identifier,
                                                          &source_location));
            REQUIRE_OK(kefir_preprocessor_token_sequence_next(mem, seq, NULL, NULL));
            matched = true;
        }

        if (!matched) {
            REQUIRE_OK(kefir_token_buffer_emplace(mem, result, current_token));
        }
    }
    return KEFIR_OK;
}

static kefir_result_t run_replacement_list_substitutions(struct kefir_mem *mem, struct kefir_preprocessor *preprocessor,
                                                         struct kefir_string_pool *symbols,
                                                         struct kefir_token_allocator *token_allocator,
                                                         const struct kefir_preprocessor_user_macro *user_macro,
                                                         const struct kefir_token_buffer *replacement,
                                                         const struct kefir_hashtree *arg_tree,
                                                         struct kefir_token_buffer *buffer) {

    struct kefir_token_buffer replacement_copy;
    REQUIRE_OK(kefir_token_buffer_init(&replacement_copy));
    kefir_result_t res = kefir_token_buffer_copy(mem, &replacement_copy, replacement);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_token_buffer_free(mem, &replacement_copy);
        return res;
    });

    struct kefir_preprocessor_token_sequence seq;
    res = kefir_preprocessor_token_sequence_init(&seq, NULL);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_token_buffer_free(mem, &replacement_copy);
        return res;
    });
    res = kefir_preprocessor_token_sequence_push_front(mem, &seq, &replacement_copy,
                                                       KEFIR_PREPROCESSOR_TOKEN_DESTINATION_NORMAL);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_preprocessor_token_sequence_free(mem, &seq);
        kefir_token_buffer_free(mem, &replacement_copy);
        return res;
    });
    res = kefir_token_buffer_free(mem, &replacement_copy);
    REQUIRE_CHAIN(&res, run_replacement_list_substitutions_impl(mem, preprocessor, symbols, token_allocator, user_macro,
                                                                &seq, arg_tree, buffer));
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_preprocessor_token_sequence_free(mem, &seq);
        return res;
    });

    REQUIRE_OK(kefir_preprocessor_token_sequence_free(mem, &seq));
    return KEFIR_OK;
}

static kefir_result_t apply_function_macro(struct kefir_mem *mem, struct kefir_preprocessor *preprocessor,
                                           const struct kefir_preprocessor_user_macro *user_macro,
                                           struct kefir_string_pool *symbols, const struct kefir_list *args,
                                           struct kefir_token_allocator *token_allocator,
                                           struct kefir_token_buffer *buffer) {
    REQUIRE(kefir_list_length(&user_macro->parameters) + (user_macro->vararg ? 1 : 0) <= kefir_list_length(args),
            KEFIR_SET_ERROR(KEFIR_OUT_OF_BOUNDS, "Argument list length mismatch"));
    struct kefir_hashtree arg_tree;
    REQUIRE_OK(kefir_hashtree_init(&arg_tree, &kefir_hashtree_str_ops));

    kefir_result_t res = build_function_macro_args_tree(mem, user_macro, args, &arg_tree);
    REQUIRE_CHAIN(&res, run_replacement_list_substitutions(mem, preprocessor, symbols, token_allocator, user_macro,
                                                           &user_macro->replacement, &arg_tree, buffer));
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_hashtree_free(mem, &arg_tree);
        return res;
    });

    REQUIRE_OK(kefir_hashtree_free(mem, &arg_tree));
    return KEFIR_OK;
}

static kefir_result_t apply_object_macro(struct kefir_mem *mem, struct kefir_preprocessor *preprocessor,
                                         const struct kefir_preprocessor_user_macro *user_macro,
                                         struct kefir_string_pool *symbols,
                                         struct kefir_token_allocator *token_allocator,
                                         struct kefir_token_buffer *buffer) {
    REQUIRE_OK(run_replacement_list_substitutions(mem, preprocessor, symbols, token_allocator, user_macro,
                                                  &user_macro->replacement, NULL, buffer));
    return KEFIR_OK;
}

static kefir_result_t user_macro_apply(struct kefir_mem *mem, struct kefir_preprocessor *preprocessor,
                                       const struct kefir_preprocessor_macro *macro, struct kefir_string_pool *symbols,
                                       const struct kefir_list *args, struct kefir_token_allocator *token_allocator,
                                       struct kefir_token_buffer *buffer,
                                       const struct kefir_source_location *source_location) {
    UNUSED(source_location);
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(preprocessor != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor"));
    REQUIRE(macro != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor macro"));
    REQUIRE(token_allocator != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid token allocator"));
    REQUIRE(buffer != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid token buffer"));
    ASSIGN_DECL_CAST(struct kefir_preprocessor_user_macro *, user_macro, macro->payload);

    struct kefir_token_buffer inner_buffer;
    kefir_result_t res;
    if (macro->type == KEFIR_PREPROCESSOR_MACRO_FUNCTION) {
        REQUIRE(args != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid macro argument list"));
        REQUIRE_OK(kefir_token_buffer_init(&inner_buffer));
        res = apply_function_macro(mem, preprocessor, user_macro, symbols, args, token_allocator, &inner_buffer);
    } else {
        REQUIRE(args == NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected empty macro argument list"));
        REQUIRE_OK(kefir_token_buffer_init(&inner_buffer));
        res = apply_object_macro(mem, preprocessor, user_macro, symbols, token_allocator, &inner_buffer);
    }
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_token_buffer_free(mem, &inner_buffer);
        return res;
    });

    const kefir_size_t inner_buffer_length = kefir_token_buffer_length(&inner_buffer);
    for (kefir_size_t i = 0; i < inner_buffer_length; i++) {
        const struct kefir_token *token = kefir_token_buffer_at(&inner_buffer, i);

        if (token->klass != KEFIR_TOKEN_PP_PLACEMAKER) {
            struct kefir_token *allocated_token;
            res = kefir_token_allocator_allocate_empty(mem, token_allocator, &allocated_token);
            REQUIRE_CHAIN(&res, kefir_token_copy(mem, allocated_token, token));
            REQUIRE_ELSE(res == KEFIR_OK, {
                kefir_token_buffer_free(mem, &inner_buffer);
                return res;
            });
            allocated_token->source_location = *source_location;

            REQUIRE_OK(kefir_token_buffer_emplace(mem, buffer, allocated_token));
        }
    }
    REQUIRE_OK(kefir_token_buffer_free(mem, &inner_buffer));

    return KEFIR_OK;
}

struct kefir_preprocessor_user_macro *kefir_preprocessor_user_macro_new_object(struct kefir_mem *mem,
                                                                               struct kefir_string_pool *symbols,
                                                                               const char *identifier) {
    REQUIRE(mem != NULL, NULL);
    REQUIRE(identifier != NULL, NULL);

    if (symbols != NULL) {
        identifier = kefir_string_pool_insert(mem, symbols, identifier, NULL);
        REQUIRE(identifier != NULL, NULL);
    }

    struct kefir_preprocessor_user_macro *macro = KEFIR_MALLOC(mem, sizeof(struct kefir_preprocessor_user_macro));
    REQUIRE(macro != NULL, NULL);

    kefir_result_t res = kefir_token_buffer_init(&macro->replacement);
    REQUIRE_ELSE(res == KEFIR_OK, {
        KEFIR_FREE(mem, macro);
        return NULL;
    });

    macro->macro.identifier = identifier;
    macro->macro.type = KEFIR_PREPROCESSOR_MACRO_OBJECT;
    macro->macro.payload = macro;
    macro->macro.argc = user_macro_argc;
    macro->macro.apply = user_macro_apply;
    return macro;
}

struct kefir_preprocessor_user_macro *kefir_preprocessor_user_macro_new_function(struct kefir_mem *mem,
                                                                                 struct kefir_string_pool *symbols,
                                                                                 const char *identifier) {
    REQUIRE(mem != NULL, NULL);
    REQUIRE(identifier != NULL, NULL);

    if (symbols != NULL) {
        identifier = kefir_string_pool_insert(mem, symbols, identifier, NULL);
        REQUIRE(identifier != NULL, NULL);
    }

    struct kefir_preprocessor_user_macro *macro = KEFIR_MALLOC(mem, sizeof(struct kefir_preprocessor_user_macro));
    REQUIRE(macro != NULL, NULL);

    kefir_result_t res = kefir_list_init(&macro->parameters);
    REQUIRE_ELSE(res == KEFIR_OK, {
        KEFIR_FREE(mem, macro);
        return NULL;
    });

    res = kefir_token_buffer_init(&macro->replacement);
    REQUIRE_ELSE(res == KEFIR_OK, {
        kefir_list_free(mem, &macro->parameters);
        KEFIR_FREE(mem, macro);
        return NULL;
    });

    macro->macro.identifier = identifier;
    macro->macro.type = KEFIR_PREPROCESSOR_MACRO_FUNCTION;
    macro->vararg = false;
    macro->vararg_parameter = NULL;
    macro->macro.payload = macro;
    macro->macro.argc = user_macro_argc;
    macro->macro.apply = user_macro_apply;
    return macro;
}

kefir_result_t kefir_preprocessor_user_macro_free(struct kefir_mem *mem, struct kefir_preprocessor_user_macro *macro) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(macro != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor macro"));

    REQUIRE_OK(kefir_token_buffer_free(mem, &macro->replacement));
    if (macro->macro.type == KEFIR_PREPROCESSOR_MACRO_FUNCTION) {
        REQUIRE_OK(kefir_list_free(mem, &macro->parameters));
    }
    macro->vararg = false;
    macro->macro.identifier = NULL;
    macro->macro.payload = NULL;
    macro->macro.argc = NULL;
    macro->macro.apply = NULL;
    KEFIR_FREE(mem, macro);
    return KEFIR_OK;
}

static kefir_result_t free_macro(struct kefir_mem *mem, struct kefir_hashtreeset *tree, kefir_hashtreeset_entry_t entry,
                                 void *payload) {
    UNUSED(tree);
    UNUSED(payload);
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    ASSIGN_DECL_CAST(struct kefir_preprocessor_user_macro *, macro, entry);
    REQUIRE(macro != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor macro"));

    REQUIRE_OK(kefir_preprocessor_user_macro_free(mem, macro));
    return KEFIR_OK;
}

static kefir_result_t user_macro_locate(const struct kefir_preprocessor_macro_scope *scope, const char *identifier,
                                        const struct kefir_preprocessor_macro **macro_ptr) {
    REQUIRE(scope != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor macro scope"));
    REQUIRE(identifier != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid identifier"));
    REQUIRE(macro_ptr != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to preprocessor macro"));

    ASSIGN_DECL_CAST(struct kefir_preprocessor_user_macro_scope *, user_macro_scope, scope->payload);
    const struct kefir_preprocessor_user_macro *user_macro;
    REQUIRE_OK(kefir_preprocessor_user_macro_scope_at(user_macro_scope, identifier, &user_macro));
    *macro_ptr = &user_macro->macro;
    return KEFIR_OK;
}

kefir_result_t kefir_preprocessor_user_macro_scope_init(const struct kefir_preprocessor_user_macro_scope *parent,
                                                        struct kefir_preprocessor_user_macro_scope *scope) {
    REQUIRE(scope != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to preprocessor macro scope"));

    REQUIRE_OK(kefir_hashtree_init(&scope->macro_index, &kefir_hashtree_str_ops));
    REQUIRE_OK(kefir_hashtreeset_init(&scope->macros, &kefir_hashtree_uint_ops));
    REQUIRE_OK(kefir_hashtreeset_on_remove(&scope->macros, free_macro, NULL));
    scope->parent = parent;
    scope->scope.locate = user_macro_locate;
    scope->scope.payload = scope;
    return KEFIR_OK;
}

kefir_result_t kefir_preprocessor_user_macro_scope_free(struct kefir_mem *mem,
                                                        struct kefir_preprocessor_user_macro_scope *scope) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(scope != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor macro scope"));

    REQUIRE_OK(kefir_hashtree_free(mem, &scope->macro_index));
    REQUIRE_OK(kefir_hashtreeset_free(mem, &scope->macros));
    return KEFIR_OK;
}

kefir_result_t kefir_preprocessor_user_macro_scope_insert(struct kefir_mem *mem,
                                                          struct kefir_preprocessor_user_macro_scope *scope,
                                                          struct kefir_preprocessor_user_macro *macro) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(scope != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor macro scope"));
    REQUIRE(macro != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor macro"));

    if (kefir_hashtree_has(&scope->macro_index, (kefir_hashtree_key_t) macro->macro.identifier)) {
        REQUIRE_OK(kefir_hashtree_delete(mem, &scope->macro_index, (kefir_hashtree_key_t) macro->macro.identifier));
    }
    REQUIRE_OK(kefir_hashtreeset_add(mem, &scope->macros, (kefir_hashtreeset_entry_t) macro));
    REQUIRE_OK(kefir_hashtree_insert(mem, &scope->macro_index, (kefir_hashtree_key_t) macro->macro.identifier,
                                     (kefir_hashtree_value_t) macro));
    return KEFIR_OK;
}

kefir_result_t kefir_preprocessor_user_macro_scope_at(const struct kefir_preprocessor_user_macro_scope *scope,
                                                      const char *identifier,
                                                      const struct kefir_preprocessor_user_macro **macro_ptr) {
    REQUIRE(scope != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor macro scope"));
    REQUIRE(identifier != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid dentifier"));
    REQUIRE(macro_ptr != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to preprocessor macro"));

    struct kefir_hashtree_node *node = NULL;
    kefir_result_t res = kefir_hashtree_at(&scope->macro_index, (kefir_hashtree_key_t) identifier, &node);
    if (res == KEFIR_NOT_FOUND && scope->parent != NULL) {
        REQUIRE_OK(kefir_preprocessor_user_macro_scope_at(scope->parent, identifier, macro_ptr));
    } else {
        REQUIRE_OK(res);
        *macro_ptr = (const struct kefir_preprocessor_user_macro *) node->value;
    }
    return KEFIR_OK;
}

kefir_bool_t kefir_preprocessor_user_macro_scope_has(const struct kefir_preprocessor_user_macro_scope *scope,
                                                     const char *identifier) {
    REQUIRE(scope != NULL, false);
    REQUIRE(identifier != NULL, false);

    return kefir_hashtree_has(&scope->macro_index, (kefir_hashtree_key_t) identifier) ||
           (scope->parent != NULL && kefir_preprocessor_user_macro_scope_has(scope->parent, identifier));
}

kefir_result_t kefir_preprocessor_user_macro_scope_remove(struct kefir_mem *mem,
                                                          struct kefir_preprocessor_user_macro_scope *scope,
                                                          const char *identifier) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(scope != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid preprocessor macro scope"));
    REQUIRE(identifier != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid dentifier"));

    kefir_result_t res = kefir_hashtree_delete(mem, &scope->macro_index, (kefir_hashtree_key_t) identifier);
    if (res != KEFIR_NOT_FOUND) {
        REQUIRE_OK(res);
    }
    return KEFIR_OK;
}
