/*
    SPDX-License-Identifier: GPL-3.0

    Copyright (C) 2020-2025  Jevgenijs Protopopovs

    This file is part of Kefir project.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "kefir/target/asm/amd64/xasmgen.h"
#include "kefir/core/error.h"
#include "kefir/core/util.h"
#include "kefir/core/hashtree.h"
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <strings.h>

kefir_result_t kefir_asm_amd64_xasmgen_register8(kefir_asm_amd64_xasmgen_register_t src,
                                                 kefir_asm_amd64_xasmgen_register_t *dst) {
    REQUIRE(dst != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to AMD64 xasmgen register"));

    switch (src) {
        case KEFIR_AMD64_XASMGEN_REGISTER_AL:
        case KEFIR_AMD64_XASMGEN_REGISTER_AX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EAX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RAX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_AL;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_AH:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_AH;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_BL:
        case KEFIR_AMD64_XASMGEN_REGISTER_BX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EBX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RBX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_BL;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_BH:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_BH;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_CL:
        case KEFIR_AMD64_XASMGEN_REGISTER_CX:
        case KEFIR_AMD64_XASMGEN_REGISTER_ECX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RCX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_CL;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_CH:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_CH;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_DL:
        case KEFIR_AMD64_XASMGEN_REGISTER_DX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EDX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RDX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_DL;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_DH:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_DH;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_SIL:
        case KEFIR_AMD64_XASMGEN_REGISTER_SI:
        case KEFIR_AMD64_XASMGEN_REGISTER_ESI:
        case KEFIR_AMD64_XASMGEN_REGISTER_RSI:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_SIL;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_DIL:
        case KEFIR_AMD64_XASMGEN_REGISTER_DI:
        case KEFIR_AMD64_XASMGEN_REGISTER_EDI:
        case KEFIR_AMD64_XASMGEN_REGISTER_RDI:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_DIL;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_SPL:
        case KEFIR_AMD64_XASMGEN_REGISTER_SP:
        case KEFIR_AMD64_XASMGEN_REGISTER_ESP:
        case KEFIR_AMD64_XASMGEN_REGISTER_RSP:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_SPL;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_BPL:
        case KEFIR_AMD64_XASMGEN_REGISTER_BP:
        case KEFIR_AMD64_XASMGEN_REGISTER_EBP:
        case KEFIR_AMD64_XASMGEN_REGISTER_RBP:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_BPL;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R8B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R8B;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R9B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R9B;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R10B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R10B;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R11B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R11B;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R12B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R12B;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R13B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R13B;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R14B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R14B;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R15B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R15B;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_XMM0:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM1:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM2:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM3:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM4:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM5:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM6:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM7:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM8:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM9:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM10:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM11:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM12:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM13:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM14:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM15:
            return KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Cannot obtain 8-bit variant of XMM register");
    }
    return KEFIR_OK;
}

kefir_result_t kefir_asm_amd64_xasmgen_register16(kefir_asm_amd64_xasmgen_register_t src,
                                                  kefir_asm_amd64_xasmgen_register_t *dst) {
    REQUIRE(dst != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to AMD64 xasmgen register"));

    switch (src) {
        case KEFIR_AMD64_XASMGEN_REGISTER_AL:
        case KEFIR_AMD64_XASMGEN_REGISTER_AH:
        case KEFIR_AMD64_XASMGEN_REGISTER_AX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EAX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RAX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_AX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_BL:
        case KEFIR_AMD64_XASMGEN_REGISTER_BH:
        case KEFIR_AMD64_XASMGEN_REGISTER_BX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EBX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RBX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_BX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_CL:
        case KEFIR_AMD64_XASMGEN_REGISTER_CH:
        case KEFIR_AMD64_XASMGEN_REGISTER_CX:
        case KEFIR_AMD64_XASMGEN_REGISTER_ECX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RCX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_CX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_DL:
        case KEFIR_AMD64_XASMGEN_REGISTER_DH:
        case KEFIR_AMD64_XASMGEN_REGISTER_DX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EDX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RDX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_DX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_SIL:
        case KEFIR_AMD64_XASMGEN_REGISTER_SI:
        case KEFIR_AMD64_XASMGEN_REGISTER_ESI:
        case KEFIR_AMD64_XASMGEN_REGISTER_RSI:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_SI;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_DIL:
        case KEFIR_AMD64_XASMGEN_REGISTER_DI:
        case KEFIR_AMD64_XASMGEN_REGISTER_EDI:
        case KEFIR_AMD64_XASMGEN_REGISTER_RDI:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_DI;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_SPL:
        case KEFIR_AMD64_XASMGEN_REGISTER_SP:
        case KEFIR_AMD64_XASMGEN_REGISTER_ESP:
        case KEFIR_AMD64_XASMGEN_REGISTER_RSP:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_SP;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_BPL:
        case KEFIR_AMD64_XASMGEN_REGISTER_BP:
        case KEFIR_AMD64_XASMGEN_REGISTER_EBP:
        case KEFIR_AMD64_XASMGEN_REGISTER_RBP:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_BP;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R8B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R8W;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R9B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R9W;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R10B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R10W;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R11B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R11W;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R12B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R12W;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R13B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R13W;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R14B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R14W;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R15B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R15W;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_XMM0:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM1:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM2:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM3:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM4:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM5:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM6:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM7:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM8:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM9:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM10:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM11:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM12:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM13:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM14:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM15:
            return KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Cannot obtain 16-bit variant of XMM register");
    }
    return KEFIR_OK;
}

kefir_result_t kefir_asm_amd64_xasmgen_register32(kefir_asm_amd64_xasmgen_register_t src,
                                                  kefir_asm_amd64_xasmgen_register_t *dst) {
    REQUIRE(dst != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to AMD64 xasmgen register"));

    switch (src) {
        case KEFIR_AMD64_XASMGEN_REGISTER_AL:
        case KEFIR_AMD64_XASMGEN_REGISTER_AH:
        case KEFIR_AMD64_XASMGEN_REGISTER_AX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EAX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RAX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_EAX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_BL:
        case KEFIR_AMD64_XASMGEN_REGISTER_BH:
        case KEFIR_AMD64_XASMGEN_REGISTER_BX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EBX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RBX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_EBX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_CL:
        case KEFIR_AMD64_XASMGEN_REGISTER_CH:
        case KEFIR_AMD64_XASMGEN_REGISTER_CX:
        case KEFIR_AMD64_XASMGEN_REGISTER_ECX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RCX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_ECX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_DL:
        case KEFIR_AMD64_XASMGEN_REGISTER_DH:
        case KEFIR_AMD64_XASMGEN_REGISTER_DX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EDX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RDX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_EDX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_SIL:
        case KEFIR_AMD64_XASMGEN_REGISTER_SI:
        case KEFIR_AMD64_XASMGEN_REGISTER_ESI:
        case KEFIR_AMD64_XASMGEN_REGISTER_RSI:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_ESI;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_DIL:
        case KEFIR_AMD64_XASMGEN_REGISTER_DI:
        case KEFIR_AMD64_XASMGEN_REGISTER_EDI:
        case KEFIR_AMD64_XASMGEN_REGISTER_RDI:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_EDI;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_SPL:
        case KEFIR_AMD64_XASMGEN_REGISTER_SP:
        case KEFIR_AMD64_XASMGEN_REGISTER_ESP:
        case KEFIR_AMD64_XASMGEN_REGISTER_RSP:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_ESP;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_BPL:
        case KEFIR_AMD64_XASMGEN_REGISTER_BP:
        case KEFIR_AMD64_XASMGEN_REGISTER_EBP:
        case KEFIR_AMD64_XASMGEN_REGISTER_RBP:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_EBP;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R8B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R8D;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R9B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R9D;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R10B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R10D;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R11B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R11D;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R12B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R12D;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R13B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R13D;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R14B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R14D;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R15B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R15D;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_XMM0:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM1:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM2:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM3:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM4:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM5:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM6:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM7:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM8:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM9:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM10:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM11:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM12:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM13:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM14:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM15:
            return KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Cannot obtain 32-bit variant of XMM register");
    }
    return KEFIR_OK;
}

kefir_result_t kefir_asm_amd64_xasmgen_register64(kefir_asm_amd64_xasmgen_register_t src,
                                                  kefir_asm_amd64_xasmgen_register_t *dst) {
    REQUIRE(dst != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to AMD64 xasmgen register"));

    switch (src) {
        case KEFIR_AMD64_XASMGEN_REGISTER_AL:
        case KEFIR_AMD64_XASMGEN_REGISTER_AH:
        case KEFIR_AMD64_XASMGEN_REGISTER_AX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EAX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RAX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_RAX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_BL:
        case KEFIR_AMD64_XASMGEN_REGISTER_BH:
        case KEFIR_AMD64_XASMGEN_REGISTER_BX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EBX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RBX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_RBX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_CL:
        case KEFIR_AMD64_XASMGEN_REGISTER_CH:
        case KEFIR_AMD64_XASMGEN_REGISTER_CX:
        case KEFIR_AMD64_XASMGEN_REGISTER_ECX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RCX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_RCX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_DL:
        case KEFIR_AMD64_XASMGEN_REGISTER_DH:
        case KEFIR_AMD64_XASMGEN_REGISTER_DX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EDX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RDX:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_RDX;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_SIL:
        case KEFIR_AMD64_XASMGEN_REGISTER_SI:
        case KEFIR_AMD64_XASMGEN_REGISTER_ESI:
        case KEFIR_AMD64_XASMGEN_REGISTER_RSI:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_RSI;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_DIL:
        case KEFIR_AMD64_XASMGEN_REGISTER_DI:
        case KEFIR_AMD64_XASMGEN_REGISTER_EDI:
        case KEFIR_AMD64_XASMGEN_REGISTER_RDI:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_RDI;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_SPL:
        case KEFIR_AMD64_XASMGEN_REGISTER_SP:
        case KEFIR_AMD64_XASMGEN_REGISTER_ESP:
        case KEFIR_AMD64_XASMGEN_REGISTER_RSP:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_RSP;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_BPL:
        case KEFIR_AMD64_XASMGEN_REGISTER_BP:
        case KEFIR_AMD64_XASMGEN_REGISTER_EBP:
        case KEFIR_AMD64_XASMGEN_REGISTER_RBP:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_RBP;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R8B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R8;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R9B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R9;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R10B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R10;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R11B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R11;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R12B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R12;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R13B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R13;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R14B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R14;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_R15B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15:
            *dst = KEFIR_AMD64_XASMGEN_REGISTER_R15;
            break;

        case KEFIR_AMD64_XASMGEN_REGISTER_XMM0:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM1:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM2:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM3:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM4:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM5:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM6:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM7:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM8:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM9:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM10:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM11:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM12:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM13:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM14:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM15:
            return KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Cannot obtain 64-bit variant of XMM register");
    }
    return KEFIR_OK;
}

kefir_bool_t kefir_asm_amd64_xasmgen_register_is_floating_point(kefir_asm_amd64_xasmgen_register_t reg) {
    switch (reg) {
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM0:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM1:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM2:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM3:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM4:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM5:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM6:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM7:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM8:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM9:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM10:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM11:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM12:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM13:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM14:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM15:
            return true;

        default:
            return false;
    }
}

kefir_bool_t kefir_asm_amd64_xasmgen_register_is_wide(kefir_asm_amd64_xasmgen_register_t reg, kefir_size_t width) {
    switch (reg) {
        case KEFIR_AMD64_XASMGEN_REGISTER_AL:
        case KEFIR_AMD64_XASMGEN_REGISTER_AH:
        case KEFIR_AMD64_XASMGEN_REGISTER_BL:
        case KEFIR_AMD64_XASMGEN_REGISTER_BH:
        case KEFIR_AMD64_XASMGEN_REGISTER_CL:
        case KEFIR_AMD64_XASMGEN_REGISTER_CH:
        case KEFIR_AMD64_XASMGEN_REGISTER_DL:
        case KEFIR_AMD64_XASMGEN_REGISTER_DH:
        case KEFIR_AMD64_XASMGEN_REGISTER_SIL:
        case KEFIR_AMD64_XASMGEN_REGISTER_DIL:
        case KEFIR_AMD64_XASMGEN_REGISTER_SPL:
        case KEFIR_AMD64_XASMGEN_REGISTER_BPL:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14B:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15B:
            return width == 8;

        case KEFIR_AMD64_XASMGEN_REGISTER_AX:
        case KEFIR_AMD64_XASMGEN_REGISTER_BX:
        case KEFIR_AMD64_XASMGEN_REGISTER_CX:
        case KEFIR_AMD64_XASMGEN_REGISTER_DX:
        case KEFIR_AMD64_XASMGEN_REGISTER_SI:
        case KEFIR_AMD64_XASMGEN_REGISTER_DI:
        case KEFIR_AMD64_XASMGEN_REGISTER_SP:
        case KEFIR_AMD64_XASMGEN_REGISTER_BP:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14W:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15W:
            return width == 16;

        case KEFIR_AMD64_XASMGEN_REGISTER_EAX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EBX:
        case KEFIR_AMD64_XASMGEN_REGISTER_ECX:
        case KEFIR_AMD64_XASMGEN_REGISTER_EDX:
        case KEFIR_AMD64_XASMGEN_REGISTER_ESI:
        case KEFIR_AMD64_XASMGEN_REGISTER_EDI:
        case KEFIR_AMD64_XASMGEN_REGISTER_ESP:
        case KEFIR_AMD64_XASMGEN_REGISTER_EBP:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14D:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15D:
            return width == 32;

        case KEFIR_AMD64_XASMGEN_REGISTER_RAX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RBX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RCX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RDX:
        case KEFIR_AMD64_XASMGEN_REGISTER_RSI:
        case KEFIR_AMD64_XASMGEN_REGISTER_RDI:
        case KEFIR_AMD64_XASMGEN_REGISTER_RSP:
        case KEFIR_AMD64_XASMGEN_REGISTER_RBP:
        case KEFIR_AMD64_XASMGEN_REGISTER_R8:
        case KEFIR_AMD64_XASMGEN_REGISTER_R9:
        case KEFIR_AMD64_XASMGEN_REGISTER_R10:
        case KEFIR_AMD64_XASMGEN_REGISTER_R11:
        case KEFIR_AMD64_XASMGEN_REGISTER_R12:
        case KEFIR_AMD64_XASMGEN_REGISTER_R13:
        case KEFIR_AMD64_XASMGEN_REGISTER_R14:
        case KEFIR_AMD64_XASMGEN_REGISTER_R15:
            return width == 64;

        case KEFIR_AMD64_XASMGEN_REGISTER_XMM0:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM1:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM2:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM3:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM4:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM5:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM6:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM7:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM8:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM9:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM10:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM11:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM12:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM13:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM14:
        case KEFIR_AMD64_XASMGEN_REGISTER_XMM15:
            return width == 128;

        default:
            return false;
    }
}

kefir_result_t kefir_asm_amd64_xasmgen_register_widest(kefir_asm_amd64_xasmgen_register_t src,
                                                       kefir_asm_amd64_xasmgen_register_t *dst) {
    REQUIRE(dst != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to AMD64 xasmgen register"));

    if (kefir_asm_amd64_xasmgen_register_is_floating_point(src)) {
        *dst = src;
    } else {
        REQUIRE_OK(kefir_asm_amd64_xasmgen_register64(src, dst));
    }
    return KEFIR_OK;
}

static const char *register_literals[] = {
    [KEFIR_AMD64_XASMGEN_REGISTER_AL] = "al",       [KEFIR_AMD64_XASMGEN_REGISTER_AH] = "ah",
    [KEFIR_AMD64_XASMGEN_REGISTER_BL] = "bl",       [KEFIR_AMD64_XASMGEN_REGISTER_BH] = "bh",
    [KEFIR_AMD64_XASMGEN_REGISTER_CL] = "cl",       [KEFIR_AMD64_XASMGEN_REGISTER_CH] = "ch",
    [KEFIR_AMD64_XASMGEN_REGISTER_DL] = "dl",       [KEFIR_AMD64_XASMGEN_REGISTER_DH] = "dh",
    [KEFIR_AMD64_XASMGEN_REGISTER_SIL] = "sil",     [KEFIR_AMD64_XASMGEN_REGISTER_DIL] = "dil",
    [KEFIR_AMD64_XASMGEN_REGISTER_SPL] = "spl",     [KEFIR_AMD64_XASMGEN_REGISTER_BPL] = "bpl",
    [KEFIR_AMD64_XASMGEN_REGISTER_R8B] = "r8b",     [KEFIR_AMD64_XASMGEN_REGISTER_R9B] = "r9b",
    [KEFIR_AMD64_XASMGEN_REGISTER_R10B] = "r10b",   [KEFIR_AMD64_XASMGEN_REGISTER_R11B] = "r11b",
    [KEFIR_AMD64_XASMGEN_REGISTER_R12B] = "r12b",   [KEFIR_AMD64_XASMGEN_REGISTER_R13B] = "r13b",
    [KEFIR_AMD64_XASMGEN_REGISTER_R14B] = "r14b",   [KEFIR_AMD64_XASMGEN_REGISTER_R15B] = "r15b",
    [KEFIR_AMD64_XASMGEN_REGISTER_AX] = "ax",       [KEFIR_AMD64_XASMGEN_REGISTER_BX] = "bx",
    [KEFIR_AMD64_XASMGEN_REGISTER_CX] = "cx",       [KEFIR_AMD64_XASMGEN_REGISTER_DX] = "dx",
    [KEFIR_AMD64_XASMGEN_REGISTER_SI] = "si",       [KEFIR_AMD64_XASMGEN_REGISTER_DI] = "di",
    [KEFIR_AMD64_XASMGEN_REGISTER_SP] = "sp",       [KEFIR_AMD64_XASMGEN_REGISTER_BP] = "bp",
    [KEFIR_AMD64_XASMGEN_REGISTER_R8W] = "r8w",     [KEFIR_AMD64_XASMGEN_REGISTER_R9W] = "r9w",
    [KEFIR_AMD64_XASMGEN_REGISTER_R10W] = "r10w",   [KEFIR_AMD64_XASMGEN_REGISTER_R11W] = "r11w",
    [KEFIR_AMD64_XASMGEN_REGISTER_R12W] = "r12w",   [KEFIR_AMD64_XASMGEN_REGISTER_R13W] = "r13w",
    [KEFIR_AMD64_XASMGEN_REGISTER_R14W] = "r14w",   [KEFIR_AMD64_XASMGEN_REGISTER_R15W] = "r15w",
    [KEFIR_AMD64_XASMGEN_REGISTER_EAX] = "eax",     [KEFIR_AMD64_XASMGEN_REGISTER_EBX] = "ebx",
    [KEFIR_AMD64_XASMGEN_REGISTER_ECX] = "ecx",     [KEFIR_AMD64_XASMGEN_REGISTER_EDX] = "edx",
    [KEFIR_AMD64_XASMGEN_REGISTER_ESI] = "esi",     [KEFIR_AMD64_XASMGEN_REGISTER_EDI] = "edi",
    [KEFIR_AMD64_XASMGEN_REGISTER_ESP] = "esp",     [KEFIR_AMD64_XASMGEN_REGISTER_EBP] = "ebp",
    [KEFIR_AMD64_XASMGEN_REGISTER_R8D] = "r8d",     [KEFIR_AMD64_XASMGEN_REGISTER_R9D] = "r9d",
    [KEFIR_AMD64_XASMGEN_REGISTER_R10D] = "r10d",   [KEFIR_AMD64_XASMGEN_REGISTER_R11D] = "r11d",
    [KEFIR_AMD64_XASMGEN_REGISTER_R12D] = "r12d",   [KEFIR_AMD64_XASMGEN_REGISTER_R13D] = "r13d",
    [KEFIR_AMD64_XASMGEN_REGISTER_R14D] = "r14d",   [KEFIR_AMD64_XASMGEN_REGISTER_R15D] = "r15d",
    [KEFIR_AMD64_XASMGEN_REGISTER_RAX] = "rax",     [KEFIR_AMD64_XASMGEN_REGISTER_RBX] = "rbx",
    [KEFIR_AMD64_XASMGEN_REGISTER_RCX] = "rcx",     [KEFIR_AMD64_XASMGEN_REGISTER_RDX] = "rdx",
    [KEFIR_AMD64_XASMGEN_REGISTER_RSI] = "rsi",     [KEFIR_AMD64_XASMGEN_REGISTER_RDI] = "rdi",
    [KEFIR_AMD64_XASMGEN_REGISTER_RSP] = "rsp",     [KEFIR_AMD64_XASMGEN_REGISTER_RBP] = "rbp",
    [KEFIR_AMD64_XASMGEN_REGISTER_R8] = "r8",       [KEFIR_AMD64_XASMGEN_REGISTER_R9] = "r9",
    [KEFIR_AMD64_XASMGEN_REGISTER_R10] = "r10",     [KEFIR_AMD64_XASMGEN_REGISTER_R11] = "r11",
    [KEFIR_AMD64_XASMGEN_REGISTER_R12] = "r12",     [KEFIR_AMD64_XASMGEN_REGISTER_R13] = "r13",
    [KEFIR_AMD64_XASMGEN_REGISTER_R14] = "r14",     [KEFIR_AMD64_XASMGEN_REGISTER_R15] = "r15",
    [KEFIR_AMD64_XASMGEN_REGISTER_XMM0] = "xmm0",   [KEFIR_AMD64_XASMGEN_REGISTER_XMM1] = "xmm1",
    [KEFIR_AMD64_XASMGEN_REGISTER_XMM2] = "xmm2",   [KEFIR_AMD64_XASMGEN_REGISTER_XMM3] = "xmm3",
    [KEFIR_AMD64_XASMGEN_REGISTER_XMM4] = "xmm4",   [KEFIR_AMD64_XASMGEN_REGISTER_XMM5] = "xmm5",
    [KEFIR_AMD64_XASMGEN_REGISTER_XMM6] = "xmm6",   [KEFIR_AMD64_XASMGEN_REGISTER_XMM7] = "xmm7",
    [KEFIR_AMD64_XASMGEN_REGISTER_XMM8] = "xmm8",   [KEFIR_AMD64_XASMGEN_REGISTER_XMM9] = "xmm9",
    [KEFIR_AMD64_XASMGEN_REGISTER_XMM10] = "xmm10", [KEFIR_AMD64_XASMGEN_REGISTER_XMM11] = "xmm11",
    [KEFIR_AMD64_XASMGEN_REGISTER_XMM12] = "xmm12", [KEFIR_AMD64_XASMGEN_REGISTER_XMM13] = "xmm13",
    [KEFIR_AMD64_XASMGEN_REGISTER_XMM14] = "xmm14", [KEFIR_AMD64_XASMGEN_REGISTER_XMM15] = "xmm15"};

static const kefir_size_t register_literal_count = sizeof(register_literals) / sizeof(register_literals[0]);

const char *kefir_asm_amd64_xasmgen_register_symbolic_name(kefir_asm_amd64_xasmgen_register_t reg) {
    return register_literals[reg];
}

kefir_result_t kefir_asm_amd64_xasmgen_register_from_symbolic_name(const char *symbolic,
                                                                   kefir_asm_amd64_xasmgen_register_t *reg_ptr) {
    REQUIRE(symbolic != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid symbolic amd64 register name"));
    REQUIRE(reg_ptr != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to AMD64 xasmgen register"));

    for (kefir_size_t i = 0; i < register_literal_count; i++) {
        if (strcmp(register_literals[i], symbolic) == 0) {
            *reg_ptr = (kefir_asm_amd64_xasmgen_register_t) i;
            return KEFIR_OK;
        }
    }
    return KEFIR_SET_ERROR(KEFIR_NOT_FOUND, "Requested symbolic amd64 register name was not found");
}

struct xasmgen_payload {
    FILE *output;
    kefir_asm_amd64_xasmgen_syntax_t syntax;
};

static kefir_result_t amd64_ident(struct kefir_amd64_xasmgen *xasmgen) {
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);
    if (xasmgen->settings.enable_identation) {
        fprintf(payload->output, "    ");
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_prologue(struct kefir_amd64_xasmgen *xasmgen) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    switch (payload->syntax) {
        case KEFIR_AMD64_XASMGEN_SYNTAX_INTEL_PREFIX:
            fprintf(payload->output, ".intel_syntax prefix\n");
            fprintf(payload->output, "%s\n\n", ".section .note.GNU-stack,\"\",%progbits");
            break;

        case KEFIR_AMD64_XASMGEN_SYNTAX_INTEL_NOPREFIX:
            fprintf(payload->output, ".intel_syntax noprefix\n");
            fprintf(payload->output, "%s\n\n", ".section .note.GNU-stack,\"\",%progbits");
            break;

        case KEFIR_AMD64_XASMGEN_SYNTAX_ATT:
            fprintf(payload->output, ".att_syntax\n");
            fprintf(payload->output, "%s\n\n", ".section .note.GNU-stack,\"\",%progbits");
            break;

        case KEFIR_AMD64_XASMGEN_SYNTAX_YASM:
            fprintf(payload->output, "bits 64\n");
            fprintf(payload->output, "%s\n\n", "section .note.GNU-stack");
            break;
    }

    return KEFIR_OK;
}

static kefir_result_t amd64_close(struct kefir_mem *mem, struct kefir_amd64_xasmgen *xasmgen) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    fflush(payload->output);
    payload->output = NULL;
    KEFIR_FREE(mem, payload);
    xasmgen->payload = NULL;
    return KEFIR_OK;
}

static kefir_result_t amd64_newline(struct kefir_amd64_xasmgen *xasmgen, unsigned int count) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    while (count--) {
        fprintf(payload->output, "\n");
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_comment(struct kefir_amd64_xasmgen *xasmgen, const char *format, ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    if (xasmgen->settings.enable_comments) {
        fprintf(payload->output, "# ");
        va_list args;
        va_start(args, format);
        vfprintf(payload->output, format, args);
        va_end(args);
        fprintf(payload->output, "\n");
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_label(struct kefir_amd64_xasmgen *xasmgen, const char *format, ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    va_list args;
    va_start(args, format);
    vfprintf(payload->output, format, args);
    va_end(args);
    fprintf(payload->output, ":\n");
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_label(struct kefir_amd64_xasmgen *xasmgen, const char *format, ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    va_list args;
    va_start(args, format);
    fprintf(payload->output, "$");
    vfprintf(payload->output, format, args);
    va_end(args);
    fprintf(payload->output, ":\n");
    return KEFIR_OK;
}

static kefir_result_t amd64_global(struct kefir_amd64_xasmgen *xasmgen, kefir_amd64_xasmgen_type_attribute_t type,
                                   kefir_amd64_xasmgen_visibility_attribute_t visibility, const char *format, ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

#define PRINT_IDENTIFIER                         \
    do {                                         \
        va_list args;                            \
        va_start(args, format);                  \
        vfprintf(payload->output, format, args); \
        va_end(args);                            \
    } while (0)

    fprintf(payload->output, ".global ");
    PRINT_IDENTIFIER;
    fprintf(payload->output, "\n");

    switch (type) {
        case KEFIR_AMD64_XASMGEN_TYPE_FUNCTION:
            fprintf(payload->output, ".type ");
            PRINT_IDENTIFIER;
            fprintf(payload->output, ", @function\n");
            break;

        case KEFIR_AMD64_XASMGEN_TYPE_DATA:
            fprintf(payload->output, ".type ");
            PRINT_IDENTIFIER;
            fprintf(payload->output, ", @object\n");
            break;
    }

    switch (visibility) {
        case KEFIR_AMD64_XASMGEN_VISIBILITY_DEFAULT:
            // Intentionally left blank
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_HIDDEN:
            fprintf(payload->output, ".hidden ");
            PRINT_IDENTIFIER;
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_INTERNAL:
            fprintf(payload->output, ".internal ");
            PRINT_IDENTIFIER;
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_PROTECTED:
            fprintf(payload->output, ".protected ");
            PRINT_IDENTIFIER;
            fprintf(payload->output, "\n");
            break;
    }
#undef PRINT_IDENTIFIER
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_global(struct kefir_amd64_xasmgen *xasmgen, kefir_amd64_xasmgen_type_attribute_t type,
                                        kefir_amd64_xasmgen_visibility_attribute_t visibility, const char *format,
                                        ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    fprintf(payload->output, "global ");
    va_list args;
    va_start(args, format);
    vfprintf(payload->output, format, args);
    va_end(args);

    switch (type) {
        case KEFIR_AMD64_XASMGEN_TYPE_FUNCTION:
            fprintf(payload->output, ":function");
            break;

        case KEFIR_AMD64_XASMGEN_TYPE_DATA:
            fprintf(payload->output, ":object");
            break;
    }

    switch (visibility) {
        case KEFIR_AMD64_XASMGEN_VISIBILITY_DEFAULT:
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_HIDDEN:
            fprintf(payload->output, " hidden\n");
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_INTERNAL:
            fprintf(payload->output, " internal\n");
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_PROTECTED:
            fprintf(payload->output, " protected\n");
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_external(struct kefir_amd64_xasmgen *xasmgen, const char *format, ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    fprintf(payload->output, ".extern ");
    va_list args;
    va_start(args, format);
    vfprintf(payload->output, format, args);
    va_end(args);
    fprintf(payload->output, "\n");
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_external(struct kefir_amd64_xasmgen *xasmgen, const char *format, ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    fprintf(payload->output, "extern ");
    va_list args;
    va_start(args, format);
    vfprintf(payload->output, format, args);
    va_end(args);
    fprintf(payload->output, "\n");
    return KEFIR_OK;
}

static kefir_result_t amd64_alias(struct kefir_amd64_xasmgen *xasmgen, const char *alias, const char *original) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    fprintf(payload->output, ".set %s, %s\n", alias, original);
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_alias(struct kefir_amd64_xasmgen *xasmgen, const char *alias, const char *original) {
    UNUSED(xasmgen);
    UNUSED(alias);
    UNUSED(original);

    return KEFIR_SET_ERROR(KEFIR_NOT_IMPLEMENTED, "Yasm assembly generator does not support aliases");
}

static kefir_result_t amd64_weak(struct kefir_amd64_xasmgen *xasmgen, kefir_amd64_xasmgen_type_attribute_t type,
                                 kefir_amd64_xasmgen_visibility_attribute_t visibility, const char *format, ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

#define PRINT_IDENTIFIER                         \
    do {                                         \
        va_list args;                            \
        va_start(args, format);                  \
        vfprintf(payload->output, format, args); \
        va_end(args);                            \
    } while (0)

    fprintf(payload->output, ".weak ");
    PRINT_IDENTIFIER;
    fprintf(payload->output, "\n");

    switch (type) {
        case KEFIR_AMD64_XASMGEN_TYPE_FUNCTION:
            fprintf(payload->output, ".type ");
            PRINT_IDENTIFIER;
            fprintf(payload->output, ", @function\n");
            break;

        case KEFIR_AMD64_XASMGEN_TYPE_DATA:
            fprintf(payload->output, ".type ");
            PRINT_IDENTIFIER;
            fprintf(payload->output, ", @object\n");
            break;
    }

    switch (visibility) {
        case KEFIR_AMD64_XASMGEN_VISIBILITY_DEFAULT:
            // Intentionally left blank
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_HIDDEN:
            fprintf(payload->output, ".hidden ");
            PRINT_IDENTIFIER;
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_INTERNAL:
            fprintf(payload->output, ".internal ");
            PRINT_IDENTIFIER;
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_PROTECTED:
            fprintf(payload->output, ".protected ");
            PRINT_IDENTIFIER;
            fprintf(payload->output, "\n");
            break;
    }

#undef PRINT_IDENTIFIER
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_weak(struct kefir_amd64_xasmgen *xasmgen, kefir_amd64_xasmgen_type_attribute_t type,
                                      kefir_amd64_xasmgen_visibility_attribute_t visibility, const char *format, ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

#define PRINT_IDENTIFIER                         \
    do {                                         \
        va_list args;                            \
        va_start(args, format);                  \
        vfprintf(payload->output, format, args); \
        va_end(args);                            \
    } while (0)

    fprintf(payload->output, "global ");
    PRINT_IDENTIFIER;

    switch (type) {
        case KEFIR_AMD64_XASMGEN_TYPE_FUNCTION:
            fprintf(payload->output, ":function");
            break;

        case KEFIR_AMD64_XASMGEN_TYPE_DATA:
            fprintf(payload->output, ":object");
            break;
    }

    switch (visibility) {
        case KEFIR_AMD64_XASMGEN_VISIBILITY_DEFAULT:
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_HIDDEN:
            fprintf(payload->output, " hidden\n");
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_INTERNAL:
            fprintf(payload->output, " internal\n");
            break;

        case KEFIR_AMD64_XASMGEN_VISIBILITY_PROTECTED:
            fprintf(payload->output, " protected\n");
            break;
    }

    fprintf(payload->output, "weak ");
    PRINT_IDENTIFIER;
    fprintf(payload->output, "\n");
    return KEFIR_OK;
}

static kefir_result_t amd64_section(struct kefir_amd64_xasmgen *xasmgen, const char *identifier,
                                    kefir_uint64_t section_attr) {
    UNUSED(section_attr);
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    fprintf(payload->output, ".section %s\n", identifier);
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_section(struct kefir_amd64_xasmgen *xasmgen, const char *identifier,
                                         kefir_uint64_t section_attr) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    if ((section_attr & KEFIR_AMD64_XASMGEN_SECTION_TLS) != 0) {
        fprintf(payload->output, "section %s tls\n", identifier);
    } else {
        fprintf(payload->output, "section %s\n", identifier);
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_align(struct kefir_amd64_xasmgen *xasmgen, kefir_size_t alignment) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    REQUIRE_OK(amd64_ident(xasmgen));
    fprintf(payload->output, ".align %" KEFIR_SIZE_FMT "\n", alignment);
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_align(struct kefir_amd64_xasmgen *xasmgen, kefir_size_t alignment) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    REQUIRE_OK(amd64_ident(xasmgen));
    fprintf(payload->output, "align %" KEFIR_SIZE_FMT "\n", alignment);
    return KEFIR_OK;
}

static kefir_result_t amd64_string_literal(void (*print)(void *, const char *, ...), void *printarg,
                                           const char *literal, kefir_size_t length) {
    print(printarg, "\"");
    const char *end = literal + length;
    for (; literal < end; ++literal) {
        switch (*literal) {
            case U'\0':
                print(printarg, "\\000");
                break;

            case U'\"':
                print(printarg, "\\\"");
                break;

            case U'\\':
                print(printarg, "\\\\");
                break;

            case U'\a':
                print(printarg, "\\%03o", '\a');
                break;

            case U'\b':
                print(printarg, "\\b");
                break;

            case U'\t':
                print(printarg, "\\t");
                break;

            case U'\n':
                print(printarg, "\\n");
                break;

            case U'\v':
                print(printarg, "\\%03o", '\v');
                break;

            case U'\f':
                print(printarg, "\\f");
                break;

            case U'\r':
                print(printarg, "\\r");
                break;

            default:
                if (isprint(*literal)) {
                    print(printarg, "%c", *literal);
                } else {
                    print(printarg, "\\%03o", (unsigned char) *literal);
                }
                break;
        }
    }
    print(printarg, "\"");
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_string_literal(void (*print)(void *, const char *, ...), void *printarg,
                                                const char *literal, kefir_size_t length) {
    const char *end = literal + length;
    for (; literal < end; ++literal) {
        print(printarg, "0x%x", *literal);
        if (literal + 1 != end) {
            print(printarg, ", ");
        }
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_symbol_arg(void (*print)(void *, const char *, ...), void *printarg,
                                       kefir_asm_amd64_xasmgen_symbol_relocation_t type, const char *symbol) {
    const char *EscapedSymbols[] = {// TODO Expand number of escaped symbols
                                    "mod"};
    kefir_bool_t printed = false;
    for (kefir_size_t i = 0; i < sizeof(EscapedSymbols) / sizeof(EscapedSymbols[0]); i++) {
        if (strcasecmp(symbol, EscapedSymbols[i]) == 0) {
            print(printarg, "$%s", symbol);
            printed = true;
            break;
        }
    }

    if (!printed) {
        print(printarg, "%s", symbol);
    }
    switch (type) {
        case KEFIR_AMD64_XASMGEN_SYMBOL_ABSOLUTE:
            // Intentionally left blank
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_PLT:
            print(printarg, "@PLT");
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_GOTPCREL:
            print(printarg, "@GOTPCREL");
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_TPOFF:
            print(printarg, "@tpoff");
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_DTPOFF:
            print(printarg, "@dtpoff");
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_GOTTPOFF:
            print(printarg, "@gottpoff");
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_TLSGD:
            print(printarg, "@tlsgd");
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_symbol_arg(void (*print)(void *, const char *, ...), void *printarg,
                                            kefir_asm_amd64_xasmgen_symbol_relocation_t type, const char *symbol) {
    print(printarg, "$%s", symbol);
    switch (type) {
        case KEFIR_AMD64_XASMGEN_SYMBOL_ABSOLUTE:
            // Intentionally left blank
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_PLT:
            print(printarg, " wrt ..plt");
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_GOTPCREL:
            print(printarg, " wrt ..gotpcrel");
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_TPOFF:
            print(printarg, " wrt ..tpoff");
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_DTPOFF:
            print(printarg, " wrt ..dtpoff");
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_GOTTPOFF:
            print(printarg, " wrt ..gottpoff");
            break;

        case KEFIR_AMD64_XASMGEN_SYMBOL_RELATIVE_TLSGD:
            print(printarg, " wrt ..tlsgd");
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t format_pointer(void (*print)(void *, const char *, ...), void *printarg,
                                     kefir_asm_amd64_xasmgen_pointer_type_t type) {
    switch (type) {
        case KEFIR_AMD64_XASMGEN_POINTER_BYTE:
            print(printarg, "BYTE PTR ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_WORD:
            print(printarg, "WORD PTR ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_DWORD:
            print(printarg, "DWORD PTR ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_QWORD:
            print(printarg, "QWORD PTR ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_TBYTE:
            print(printarg, "TBYTE PTR ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_XMMWORD:
            print(printarg, "XMMWORD PTR ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_FP_SINGLE:
            print(printarg, "DWORD PTR ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_FP_DOUBLE:
            print(printarg, "QWORD PTR ");
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t format_yasm_pointer(void (*print)(void *, const char *, ...), void *printarg,
                                          kefir_asm_amd64_xasmgen_pointer_type_t type) {
    switch (type) {
        case KEFIR_AMD64_XASMGEN_POINTER_BYTE:
            print(printarg, "BYTE ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_WORD:
            print(printarg, "WORD ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_DWORD:
            print(printarg, "DWORD ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_QWORD:
            print(printarg, "QWORD ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_TBYTE:
            print(printarg, "TWORD ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_XMMWORD:
            print(printarg, "OWORD ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_FP_SINGLE:
            print(printarg, "DWORD ");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_FP_DOUBLE:
            print(printarg, "QWORD ");
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_format_operand_intel(void (*print)(void *, const char *, ...), void *printarg,
                                                 kefir_bool_t prefix,
                                                 const struct kefir_asm_amd64_xasmgen_operand *op) {
    switch (op->klass) {
        case KEFIR_AMD64_XASMGEN_OPERAND_IMMEDIATE:
            print(printarg, "%" KEFIR_INT64_FMT, op->imm);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_IMMEDIATE_UNSIGNED:
            print(printarg, "%" KEFIR_UINT64_FMT, op->immu);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_REGISTER:
            if (prefix) {
                print(printarg, "%%%s", register_literals[op->reg]);
            } else {
                print(printarg, "%s", register_literals[op->reg]);
            }
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_LABEL:
            REQUIRE_OK(amd64_symbol_arg(print, printarg, op->label.type, op->label.symbol));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_INDIRECTION:
            print(printarg, "[");
            REQUIRE_OK(amd64_format_operand_intel(print, printarg, prefix, op->indirection.base));
            if (op->indirection.displacement > 0) {
                print(printarg, " + %" KEFIR_INT64_FMT, op->indirection.displacement);
            } else if (op->indirection.displacement < 0) {
                print(printarg, " - %" KEFIR_INT64_FMT, -op->indirection.displacement);
            }
            print(printarg, "]");
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_OFFSET:
            REQUIRE_OK(amd64_format_operand_intel(print, printarg, prefix, op->offset.base));
            if (op->offset.offset > 0) {
                print(printarg, " + %" KEFIR_INT64_FMT, op->offset.offset);
            } else if (op->offset.offset < 0) {
                print(printarg, " - %" KEFIR_INT64_FMT, -op->offset.offset);
            }
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_SEGMENT:
            switch (op->segment.segment) {
                case KEFIR_AMD64_XASMGEN_SEGMENT_FS:
                    if (prefix) {
                        print(printarg, "%%fs:");
                    } else {
                        print(printarg, "fs:");
                    }
                    break;
            }
            REQUIRE_OK(amd64_format_operand_intel(print, printarg, prefix, op->segment.base));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_POINTER:
            REQUIRE_OK(format_pointer(print, printarg, op->pointer.type));
            REQUIRE_OK(amd64_format_operand_intel(print, printarg, prefix, op->pointer.base));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_RIP_INDIRECTION:
            REQUIRE_OK(amd64_symbol_arg(print, printarg, op->label.type, op->label.symbol));
            if (prefix) {
                print(printarg, "[%%rip]");
            } else {
                print(printarg, "[rip]");
            }
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_STRING_LITERAL:
            REQUIRE_OK(amd64_string_literal(print, printarg, op->string_literal.content, op->string_literal.length));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_FPU_REGISTER:
            if (prefix) {
                print(printarg, "%%st(%" KEFIR_UINT64_FMT ")", op->fpu_register);
            } else {
                print(printarg, "st(%" KEFIR_UINT64_FMT ")", op->fpu_register);
            }
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_SUBTRACT:
            REQUIRE_OK(amd64_format_operand_intel(print, printarg, prefix, op->arithmetics.args[0]));
            print(printarg, " - ");
            REQUIRE_OK(amd64_format_operand_intel(print, printarg, prefix, op->arithmetics.args[1]));
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_format_operand_yasm(void (*print)(void *, const char *, ...), void *printarg,
                                                const struct kefir_asm_amd64_xasmgen_operand *op) {
    switch (op->klass) {
        case KEFIR_AMD64_XASMGEN_OPERAND_IMMEDIATE:
            print(printarg, "%" KEFIR_INT64_FMT, op->imm);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_IMMEDIATE_UNSIGNED:
            print(printarg, "%" KEFIR_UINT64_FMT, op->immu);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_REGISTER:
            print(printarg, "%s", register_literals[op->reg]);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_LABEL:
            REQUIRE_OK(amd64_yasm_symbol_arg(print, printarg, op->label.type, op->label.symbol));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_INDIRECTION:
            print(printarg, "[");
            REQUIRE_OK(amd64_format_operand_yasm(print, printarg, op->indirection.base));
            if (op->indirection.displacement > 0) {
                print(printarg, " + %" KEFIR_INT64_FMT, op->indirection.displacement);
            } else if (op->indirection.displacement < 0) {
                print(printarg, " - %" KEFIR_INT64_FMT, -op->indirection.displacement);
            }
            print(printarg, "]");
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_OFFSET:
            REQUIRE_OK(amd64_format_operand_yasm(print, printarg, op->offset.base));
            if (op->offset.offset > 0) {
                print(printarg, " + %" KEFIR_INT64_FMT, op->offset.offset);
            } else if (op->offset.offset < 0) {
                print(printarg, " - %" KEFIR_INT64_FMT, -op->offset.offset);
            }
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_SEGMENT:
            switch (op->segment.segment) {
                case KEFIR_AMD64_XASMGEN_SEGMENT_FS:
                    print(printarg, "[fs:");
                    break;
            }
            REQUIRE_OK(amd64_format_operand_yasm(print, printarg, op->segment.base));
            print(printarg, "]");
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_POINTER:
            REQUIRE_OK(format_yasm_pointer(print, printarg, op->pointer.type));
            REQUIRE_OK(amd64_format_operand_yasm(print, printarg, op->pointer.base));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_RIP_INDIRECTION:
            print(printarg, "[rel ");
            REQUIRE_OK(amd64_yasm_symbol_arg(print, printarg, op->label.type, op->label.symbol));
            print(printarg, "]");
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_STRING_LITERAL:
            REQUIRE_OK(
                amd64_yasm_string_literal(print, printarg, op->string_literal.content, op->string_literal.length));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_FPU_REGISTER:
            print(printarg, "st%" KEFIR_UINT64_FMT "", op->fpu_register);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_SUBTRACT:
            REQUIRE_OK(amd64_format_operand_yasm(print, printarg, op->arithmetics.args[0]));
            print(printarg, " - ");
            REQUIRE_OK(amd64_format_operand_yasm(print, printarg, op->arithmetics.args[1]));
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_format_unprefixed_operand_att(void (*)(void *, const char *, ...), void *,
                                                          const struct kefir_asm_amd64_xasmgen_operand *);

static kefir_result_t amd64_format_operand_att(void (*print)(void *, const char *, ...), void *printarg,
                                               const struct kefir_asm_amd64_xasmgen_operand *op) {
    switch (op->klass) {
        case KEFIR_AMD64_XASMGEN_OPERAND_IMMEDIATE:
            print(printarg, "$%" KEFIR_INT64_FMT, op->imm);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_IMMEDIATE_UNSIGNED:
            print(printarg, "$%" KEFIR_UINT64_FMT, op->immu);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_REGISTER:
            print(printarg, "%%%s", register_literals[op->reg]);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_LABEL:
            REQUIRE_OK(amd64_symbol_arg(print, printarg, op->label.type, op->label.symbol));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_INDIRECTION:
            if (op->indirection.displacement != 0) {
                print(printarg, "%" KEFIR_INT64_FMT, op->indirection.displacement);
            }
            print(printarg, "(");
            REQUIRE_OK(amd64_format_operand_att(print, printarg, op->indirection.base));
            print(printarg, ")");
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_OFFSET:
            REQUIRE_OK(amd64_format_operand_att(print, printarg, op->offset.base));
            if (op->offset.offset > 0) {
                print(printarg, " + %" KEFIR_INT64_FMT, op->offset.offset);
            } else if (op->offset.offset < 0) {
                print(printarg, " - %" KEFIR_INT64_FMT, -op->offset.offset);
            }
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_SEGMENT:
            switch (op->segment.segment) {
                case KEFIR_AMD64_XASMGEN_SEGMENT_FS:
                    print(printarg, "%%fs:");
                    break;
            }
            REQUIRE_OK(amd64_format_unprefixed_operand_att(print, printarg, op->segment.base));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_POINTER:
            REQUIRE_OK(amd64_format_operand_att(print, printarg, op->pointer.base));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_RIP_INDIRECTION:
            REQUIRE_OK(amd64_symbol_arg(print, printarg, op->label.type, op->label.symbol));
            print(printarg, "(%%rip)");
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_STRING_LITERAL:
            REQUIRE_OK(amd64_string_literal(print, printarg, op->string_literal.content, op->string_literal.length));
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_FPU_REGISTER:
            print(printarg, "%%st(%" KEFIR_UINT64_FMT ")", op->fpu_register);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_SUBTRACT:
            REQUIRE_OK(amd64_format_operand_att(print, printarg, op->arithmetics.args[0]));
            print(printarg, " - ");
            REQUIRE_OK(amd64_format_operand_att(print, printarg, op->arithmetics.args[1]));
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_format_unprefixed_operand_att(void (*print)(void *, const char *, ...), void *printarg,
                                                          const struct kefir_asm_amd64_xasmgen_operand *op) {
    switch (op->klass) {
        case KEFIR_AMD64_XASMGEN_OPERAND_IMMEDIATE:
            print(printarg, "%" KEFIR_INT64_FMT, op->imm);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_IMMEDIATE_UNSIGNED:
            print(printarg, "%" KEFIR_UINT64_FMT, op->immu);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_REGISTER:
            print(printarg, "%s", register_literals[op->reg]);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_FPU_REGISTER:
            print(printarg, "st(%" KEFIR_UINT64_FMT ")", op->fpu_register);
            break;

        case KEFIR_AMD64_XASMGEN_OPERAND_SUBTRACT:
            REQUIRE_OK(amd64_format_unprefixed_operand_att(print, printarg, op->arithmetics.args[0]));
            print(printarg, " - ");
            REQUIRE_OK(amd64_format_unprefixed_operand_att(print, printarg, op->arithmetics.args[1]));
            break;

        default:
            REQUIRE_OK(amd64_format_operand_att(print, printarg, op));
            break;
    }
    return KEFIR_OK;
}

static void print_fprintf(void *arg, const char *fmt, ...) {
    va_list args;
    va_start(args, fmt);
    vfprintf((FILE *) arg, fmt, args);
    va_end(args);
}

static kefir_result_t amd64_data(struct kefir_amd64_xasmgen *xasmgen, kefir_asm_amd64_xasmgen_data_type_t type,
                                 kefir_size_t length, ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    REQUIRE_OK(amd64_ident(xasmgen));
    switch (type) {
        case KEFIR_AMD64_XASMGEN_DATA_BYTE:
            fprintf(payload->output, ".byte ");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_WORD:
            fprintf(payload->output, ".word ");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_DOUBLE:
            fprintf(payload->output, ".long ");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_QUAD:
            fprintf(payload->output, ".quad ");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_ASCII:
            fprintf(payload->output, ".ascii ");
            break;
    }

    va_list args;
    va_start(args, length);
    while (length--) {
        const struct kefir_asm_amd64_xasmgen_operand *op = va_arg(args, const struct kefir_asm_amd64_xasmgen_operand *);
        if (payload->syntax != KEFIR_AMD64_XASMGEN_SYNTAX_ATT) {
            REQUIRE_OK(amd64_format_operand_intel(print_fprintf, payload->output,
                                                  payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_INTEL_PREFIX, op));
        } else {
            REQUIRE_OK(amd64_format_unprefixed_operand_att(print_fprintf, payload->output, op));
        }
        if (length > 0) {
            fprintf(payload->output, ", ");
        }
    }
    va_end(args);
    fprintf(payload->output, "\n");
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_data(struct kefir_amd64_xasmgen *xasmgen, kefir_asm_amd64_xasmgen_data_type_t type,
                                      kefir_size_t length, ...) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    REQUIRE_OK(amd64_ident(xasmgen));
    switch (type) {
        case KEFIR_AMD64_XASMGEN_DATA_BYTE:
            fprintf(payload->output, "db ");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_WORD:
            fprintf(payload->output, "dw ");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_DOUBLE:
            fprintf(payload->output, "dd ");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_QUAD:
            fprintf(payload->output, "dq ");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_ASCII:
            fprintf(payload->output, "db ");
            break;
    }

    va_list args;
    va_start(args, length);
    while (length--) {
        const struct kefir_asm_amd64_xasmgen_operand *op = va_arg(args, const struct kefir_asm_amd64_xasmgen_operand *);
        REQUIRE_OK(amd64_format_operand_yasm(print_fprintf, payload->output, op));
        if (length > 0) {
            fprintf(payload->output, ", ");
        }
    }
    va_end(args);
    fprintf(payload->output, "\n");
    return KEFIR_OK;
}

static kefir_result_t amd64_zerodata(struct kefir_amd64_xasmgen *xasmgen, kefir_size_t length) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    REQUIRE_OK(amd64_ident(xasmgen));
    fprintf(payload->output, ".zero %" KEFIR_SIZE_FMT "\n", length);
    return KEFIR_OK;
}

static kefir_result_t amd64_uninitdata(struct kefir_amd64_xasmgen *xasmgen, kefir_size_t length) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    REQUIRE_OK(amd64_ident(xasmgen));
    fprintf(payload->output, ".skip %" KEFIR_SIZE_FMT "\n", length);
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_zerodata(struct kefir_amd64_xasmgen *xasmgen, kefir_size_t length) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    REQUIRE_OK(amd64_ident(xasmgen));
    fprintf(payload->output, "times %" KEFIR_SIZE_FMT " db 0\n", length);
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_uninitdata(struct kefir_amd64_xasmgen *xasmgen, kefir_size_t length) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    REQUIRE_OK(amd64_ident(xasmgen));
    fprintf(payload->output, "resb %" KEFIR_SIZE_FMT "\n", length);
    return KEFIR_OK;
}

static kefir_result_t amd64_bindata(struct kefir_amd64_xasmgen *xasmgen, kefir_asm_amd64_xasmgen_data_type_t type,
                                    const void *ptr, kefir_size_t length) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    REQUIRE(ptr != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid binary data"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    REQUIRE_OK(amd64_ident(xasmgen));
    switch (type) {
        case KEFIR_AMD64_XASMGEN_DATA_BYTE:
            fprintf(payload->output, ".byte ");
            for (kefir_size_t i = 0; i < length; i++) {
                fprintf(payload->output, "0x%02x", ((const kefir_uint8_t *) ptr)[i]);
                if (i + 1 < length) {
                    fprintf(payload->output, ", ");
                }
            }
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_WORD:
            fprintf(payload->output, ".word ");
            for (kefir_size_t i = 0; i < length; i++) {
                fprintf(payload->output, "0x%04x", ((const kefir_uint16_t *) ptr)[i]);
                if (i + 1 < length) {
                    fprintf(payload->output, ", ");
                }
            }
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_DOUBLE:
            fprintf(payload->output, ".long ");
            for (kefir_size_t i = 0; i < length; i++) {
                fprintf(payload->output, "0x%08x", ((const kefir_uint32_t *) ptr)[i]);
                if (i + 1 < length) {
                    fprintf(payload->output, ", ");
                }
            }
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_QUAD:
            fprintf(payload->output, ".quad ");
            for (kefir_size_t i = 0; i < length; i++) {
                fprintf(payload->output, "0x%016llx", (kefir_ulong_long_t) ((const kefir_uint64_t *) ptr)[i]);
                if (i + 1 < length) {
                    fprintf(payload->output, ", ");
                }
            }
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_ASCII:
            REQUIRE_OK(amd64_string_literal(print_fprintf, payload->output, ptr, length));
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_yasm_bindata(struct kefir_amd64_xasmgen *xasmgen, kefir_asm_amd64_xasmgen_data_type_t type,
                                         const void *ptr, kefir_size_t length) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    REQUIRE(ptr != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid binary data"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    REQUIRE_OK(amd64_ident(xasmgen));
    switch (type) {
        case KEFIR_AMD64_XASMGEN_DATA_BYTE:
            fprintf(payload->output, "db ");
            for (kefir_size_t i = 0; i < length; i++) {
                fprintf(payload->output, "0x%02x", ((const kefir_uint8_t *) ptr)[i]);
                if (i + 1 < length) {
                    fprintf(payload->output, ", ");
                }
            }
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_WORD:
            fprintf(payload->output, "dw ");
            for (kefir_size_t i = 0; i < length; i++) {
                fprintf(payload->output, "0x%04x", ((const kefir_uint16_t *) ptr)[i]);
                if (i + 1 < length) {
                    fprintf(payload->output, ", ");
                }
            }
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_DOUBLE:
            fprintf(payload->output, "dd ");
            for (kefir_size_t i = 0; i < length; i++) {
                fprintf(payload->output, "0x%08x", ((const kefir_uint32_t *) ptr)[i]);
                if (i + 1 < length) {
                    fprintf(payload->output, ", ");
                }
            }
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_QUAD:
            fprintf(payload->output, "dq ");
            for (kefir_size_t i = 0; i < length; i++) {
                fprintf(payload->output, "0x%016llx", (kefir_ulong_long_t) ((const kefir_uint64_t *) ptr)[i]);
                if (i + 1 < length) {
                    fprintf(payload->output, ", ");
                }
            }
            fprintf(payload->output, "\n");
            break;

        case KEFIR_AMD64_XASMGEN_DATA_ASCII:
            REQUIRE_OK(amd64_string_literal(print_fprintf, payload->output, ptr, length));
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_inline_assembly(struct kefir_amd64_xasmgen *xasmgen, const char *text) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    REQUIRE(text != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid inline assembly text"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    fprintf(payload->output, "%s\n", text);
    return KEFIR_OK;
}

struct op_to_str {
    char *buffer;
    kefir_size_t length;
};

static void print_op_to_str(void *arg, const char *fmt, ...) {
    ASSIGN_DECL_CAST(struct op_to_str *, op_to_str, arg);
    va_list args;
    va_start(args, fmt);
    int len = vsnprintf(op_to_str->buffer, op_to_str->length, fmt, args);
    op_to_str->length -= len;
    op_to_str->buffer += len;
    va_end(args);
}

static kefir_result_t amd64_format_operand(struct kefir_amd64_xasmgen *xasmgen,
                                           const struct kefir_asm_amd64_xasmgen_operand *op, char *buf,
                                           kefir_size_t buflen) {
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator"));
    REQUIRE(buflen > 0, KEFIR_SET_ERROR(KEFIR_OUT_OF_BOUNDS, "Expected non-zero buffer length"));
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    struct op_to_str op_to_str_buf = {.buffer = buf, .length = buflen};
    if (payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_ATT) {
        REQUIRE_OK(amd64_format_operand_att(print_op_to_str, &op_to_str_buf, op));
    } else if (payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_YASM) {
        REQUIRE_OK(amd64_format_operand_yasm(print_op_to_str, &op_to_str_buf, op));
    } else {
        REQUIRE_OK(amd64_format_operand_intel(print_op_to_str, &op_to_str_buf,
                                              payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_INTEL_PREFIX, op));
    }
    return KEFIR_OK;
}

static kefir_result_t amd64_format_op(struct kefir_amd64_xasmgen *xasmgen,
                                      const struct kefir_asm_amd64_xasmgen_operand *op) {
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    if (payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_ATT) {
        REQUIRE_OK(amd64_format_operand_att(print_fprintf, payload->output, op));
    } else if (payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_YASM) {
        REQUIRE_OK(amd64_format_operand_yasm(print_fprintf, payload->output, op));
    } else {
        REQUIRE_OK(amd64_format_operand_intel(print_fprintf, payload->output,
                                              payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_INTEL_PREFIX, op));
    }
    return KEFIR_OK;
}

static kefir_result_t format_mnemonic(struct kefir_amd64_xasmgen *xasmgen, const char *mnemonic) {
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);
    // See: https://sourceware.org/binutils/docs/as/i386_002dMnemonics.html#AT_0026T-Mnemonic-versus-Intel-Mnemonic
    if (payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_ATT) {
        fprintf(payload->output, "%s", mnemonic);
    } else if (payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_YASM) {
        if (strcmp(mnemonic, "movabs") == 0) {
            fprintf(payload->output, "mov");
        } else if (strcmp(mnemonic, "movsl") == 0) {
            fprintf(payload->output, "movsd");
        } else if (strcmp(mnemonic, "data16") == 0) {
            fprintf(payload->output, "db 0x66\n");
        } else if (strcmp(mnemonic, "rex.W") == 0) {
            fprintf(payload->output, "db 0x48\n");
        } else {
            fprintf(payload->output, "%s", mnemonic);
        }
    } else if (strcmp(mnemonic, "fsubp") == 0) {
        fprintf(payload->output, "fsubrp");
    } else if (strcmp(mnemonic, "fdivp") == 0) {
        fprintf(payload->output, "fdivrp");
    } else if (strcmp(mnemonic, "fsubrp") == 0) {
        fprintf(payload->output, "fsubp");
    } else if (strcmp(mnemonic, "fdivrp") == 0) {
        fprintf(payload->output, "fdivp");
    } else if (strcmp(mnemonic, "movsl") == 0) {
        fprintf(payload->output, "movsd");
    } else {
        fprintf(payload->output, "%s", mnemonic);
    }
    return KEFIR_OK;
}

static kefir_result_t format_intel_mnemonic_suffix(struct kefir_amd64_xasmgen *xasmgen, const char *mnemonic,
                                                   const struct kefir_asm_amd64_xasmgen_operand *op1,
                                                   const struct kefir_asm_amd64_xasmgen_operand *op2,
                                                   const struct kefir_asm_amd64_xasmgen_operand *op3) {
    UNUSED(op3);
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);
    if (payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_YASM &&
        (strcmp(mnemonic, "movsx") == 0 || strcmp(mnemonic, "movzx") == 0) && op1 != NULL &&
        op1->klass == KEFIR_AMD64_XASMGEN_OPERAND_REGISTER && kefir_asm_amd64_xasmgen_register_is_wide(op1->reg, 64) &&
        op2 != NULL &&
        ((op2->klass == KEFIR_AMD64_XASMGEN_OPERAND_REGISTER &&
          kefir_asm_amd64_xasmgen_register_is_wide(op2->reg, 32)) ||
         (op2->klass == KEFIR_AMD64_XASMGEN_OPERAND_POINTER &&
          op2->pointer.type == KEFIR_AMD64_XASMGEN_POINTER_DWORD))) {
        fprintf(payload->output, "d");
    }
    return KEFIR_OK;
}

#define INSTR0_INTEL(_id, _mnemonic)                                                                                   \
    static kefir_result_t amd64_instr_intel_##_id(struct kefir_amd64_xasmgen *xasmgen) {                               \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        REQUIRE_OK(format_intel_mnemonic_suffix(xasmgen, _mnemonic, NULL, NULL, NULL));                                \
        fprintf(payload->output, "\n");                                                                                \
        return KEFIR_OK;                                                                                               \
    }

#define INSTR1_INTEL(_id, _mnemonic)                                                                                   \
    static kefir_result_t amd64_instr_intel_##_id(struct kefir_amd64_xasmgen *xasmgen,                                 \
                                                  const struct kefir_asm_amd64_xasmgen_operand *op1) {                 \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        REQUIRE(op1 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        REQUIRE_OK(format_intel_mnemonic_suffix(xasmgen, _mnemonic, op1, NULL, NULL));                                 \
        fprintf(payload->output, " ");                                                                                 \
        REQUIRE_OK(amd64_format_op(xasmgen, op1));                                                                     \
        fprintf(payload->output, "\n");                                                                                \
        return KEFIR_OK;                                                                                               \
    }

#define INSTR2_INTEL(_id, _mnemonic)                                                                                   \
    static kefir_result_t amd64_instr_intel_##_id(struct kefir_amd64_xasmgen *xasmgen,                                 \
                                                  const struct kefir_asm_amd64_xasmgen_operand *op1,                   \
                                                  const struct kefir_asm_amd64_xasmgen_operand *op2) {                 \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        REQUIRE(op1 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        REQUIRE(op2 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        REQUIRE_OK(format_intel_mnemonic_suffix(xasmgen, _mnemonic, op1, op2, NULL));                                  \
        fprintf(payload->output, " ");                                                                                 \
        REQUIRE_OK(amd64_format_op(xasmgen, op1));                                                                     \
        fprintf(payload->output, ", ");                                                                                \
        REQUIRE_OK(amd64_format_op(xasmgen, op2));                                                                     \
        fprintf(payload->output, "\n");                                                                                \
        return KEFIR_OK;                                                                                               \
    }

#define INSTR3_INTEL(_id, _mnemonic)                                                                                   \
    static kefir_result_t amd64_instr_intel_##_id(                                                                     \
        struct kefir_amd64_xasmgen *xasmgen, const struct kefir_asm_amd64_xasmgen_operand *op1,                        \
        const struct kefir_asm_amd64_xasmgen_operand *op2, const struct kefir_asm_amd64_xasmgen_operand *op3) {        \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        REQUIRE(op1 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        REQUIRE(op2 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        REQUIRE(op3 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        REQUIRE_OK(format_intel_mnemonic_suffix(xasmgen, _mnemonic, op1, op2, op3));                                   \
        fprintf(payload->output, " ");                                                                                 \
        REQUIRE_OK(amd64_format_op(xasmgen, op1));                                                                     \
        fprintf(payload->output, ", ");                                                                                \
        REQUIRE_OK(amd64_format_op(xasmgen, op2));                                                                     \
        fprintf(payload->output, ", ");                                                                                \
        REQUIRE_OK(amd64_format_op(xasmgen, op3));                                                                     \
        fprintf(payload->output, "\n");                                                                                \
        return KEFIR_OK;                                                                                               \
    }

static kefir_result_t format_att_mnemonic_suffix_impl(FILE *output, kefir_asm_amd64_xasmgen_pointer_type_t ptr) {
    switch (ptr) {
        case KEFIR_AMD64_XASMGEN_POINTER_BYTE:
            fprintf(output, "b");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_WORD:
            fprintf(output, "w");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_DWORD:
        case KEFIR_AMD64_XASMGEN_POINTER_FP_DOUBLE:
            fprintf(output, "l");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_QWORD:
            fprintf(output, "q");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_TBYTE:
            fprintf(output, "t");
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_XMMWORD:
            // Intentionally left blank
            break;

        case KEFIR_AMD64_XASMGEN_POINTER_FP_SINGLE:
            fprintf(output, "s");
            break;
    }
    return KEFIR_OK;
}

static kefir_result_t format_att_mnemonic_suffix(struct kefir_amd64_xasmgen *xasmgen,
                                                 const struct kefir_asm_amd64_xasmgen_operand *op1,
                                                 const struct kefir_asm_amd64_xasmgen_operand *op2,
                                                 const struct kefir_asm_amd64_xasmgen_operand *op3) {
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);
    if (op1 != NULL && op1->klass == KEFIR_AMD64_XASMGEN_OPERAND_POINTER) {
        REQUIRE_OK(format_att_mnemonic_suffix_impl(payload->output, op1->pointer.type));
    } else if (op2 != NULL && op2->klass == KEFIR_AMD64_XASMGEN_OPERAND_POINTER) {
        REQUIRE_OK(format_att_mnemonic_suffix_impl(payload->output, op2->pointer.type));
    } else if (op3 != NULL && op3->klass == KEFIR_AMD64_XASMGEN_OPERAND_POINTER) {
        REQUIRE_OK(format_att_mnemonic_suffix_impl(payload->output, op3->pointer.type));
    }
    fprintf(payload->output, " ");
    return KEFIR_OK;
}

#define INSTR0_ATT(_id, _mnemonic)                                                                                     \
    static kefir_result_t amd64_instr_att_##_id(struct kefir_amd64_xasmgen *xasmgen) {                                 \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        fprintf(payload->output, "\n");                                                                                \
        return KEFIR_OK;                                                                                               \
    }

#define INSTR1_ATT(_id, _mnemonic)                                                                                     \
    static kefir_result_t amd64_instr_att_##_id(struct kefir_amd64_xasmgen *xasmgen,                                   \
                                                const struct kefir_asm_amd64_xasmgen_operand *op1) {                   \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        REQUIRE(op1 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        REQUIRE_OK(format_att_mnemonic_suffix(xasmgen, op1, NULL, NULL));                                              \
        REQUIRE_OK(amd64_format_op(xasmgen, op1));                                                                     \
        fprintf(payload->output, "\n");                                                                                \
        return KEFIR_OK;                                                                                               \
    }

#define INSTR1_ATT_BR(_id, _mnemonic)                                                                                  \
    static kefir_result_t amd64_instr_att_##_id(struct kefir_amd64_xasmgen *xasmgen,                                   \
                                                const struct kefir_asm_amd64_xasmgen_operand *op1) {                   \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        REQUIRE(op1 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        REQUIRE_OK(format_att_mnemonic_suffix(xasmgen, op1, NULL, NULL));                                              \
        if (op1->klass != KEFIR_AMD64_XASMGEN_OPERAND_LABEL) {                                                         \
            fprintf(payload->output, "*");                                                                             \
        }                                                                                                              \
        REQUIRE_OK(amd64_format_op(xasmgen, op1));                                                                     \
        fprintf(payload->output, "\n");                                                                                \
        return KEFIR_OK;                                                                                               \
    }

#define INSTR2_ATT(_id, _mnemonic)                                                                                     \
    static kefir_result_t amd64_instr_att_##_id(struct kefir_amd64_xasmgen *xasmgen,                                   \
                                                const struct kefir_asm_amd64_xasmgen_operand *op1,                     \
                                                const struct kefir_asm_amd64_xasmgen_operand *op2) {                   \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        REQUIRE(op1 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        REQUIRE(op2 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        REQUIRE_OK(format_att_mnemonic_suffix(xasmgen, op1, op2, NULL));                                               \
        REQUIRE_OK(amd64_format_op(xasmgen, op2));                                                                     \
        fprintf(payload->output, ", ");                                                                                \
        REQUIRE_OK(amd64_format_op(xasmgen, op1));                                                                     \
        fprintf(payload->output, "\n");                                                                                \
        return KEFIR_OK;                                                                                               \
    }

#define INSTR3_ATT(_id, _mnemonic)                                                                                     \
    static kefir_result_t amd64_instr_att_##_id(                                                                       \
        struct kefir_amd64_xasmgen *xasmgen, const struct kefir_asm_amd64_xasmgen_operand *op1,                        \
        const struct kefir_asm_amd64_xasmgen_operand *op2, const struct kefir_asm_amd64_xasmgen_operand *op3) {        \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        REQUIRE(op1 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        REQUIRE(op2 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        REQUIRE(op3 != NULL,                                                                                           \
                KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator operand"));          \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        REQUIRE_OK(format_att_mnemonic_suffix(xasmgen, op1, op2, op3));                                                \
        REQUIRE_OK(amd64_format_op(xasmgen, op3));                                                                     \
        fprintf(payload->output, ", ");                                                                                \
        REQUIRE_OK(amd64_format_op(xasmgen, op2));                                                                     \
        fprintf(payload->output, ", ");                                                                                \
        REQUIRE_OK(amd64_format_op(xasmgen, op1));                                                                     \
        fprintf(payload->output, "\n");                                                                                \
        return KEFIR_OK;                                                                                               \
    }

#define INSTR_REPEATED0(_id, _mnemonic)                                                                                \
    static kefir_result_t amd64_instr_##_id(struct kefir_amd64_xasmgen *xasmgen, kefir_bool_t rep) {                   \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
                                                                                                                       \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        if (rep) {                                                                                                     \
            fprintf(payload->output, "rep ");                                                                          \
        }                                                                                                              \
                                                                                                                       \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        fprintf(payload->output, "\n");                                                                                \
        return KEFIR_OK;                                                                                               \
    }

#define INSTR_PREFIX0(_id, _type, _mnemonic)                                                                           \
    static kefir_result_t amd64_instr_##_type##_##_id(struct kefir_amd64_xasmgen *xasmgen) {                           \
        REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid AMD64 assembly generator")); \
        ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);                                         \
        REQUIRE_OK(amd64_ident(xasmgen));                                                                              \
        REQUIRE_OK(format_mnemonic(xasmgen, _mnemonic));                                                               \
        fprintf(payload->output, " ");                                                                                 \
        return KEFIR_OK;                                                                                               \
    }

#define INSTR0(_id, _mnemonic) \
    INSTR0_ATT(_id, _mnemonic) \
    INSTR0_INTEL(_id, _mnemonic)

#define INSTR1(_id, _mnemonic) \
    INSTR1_ATT(_id, _mnemonic) \
    INSTR1_INTEL(_id, _mnemonic)

#define INSTR2(_id, _mnemonic) \
    INSTR2_ATT(_id, _mnemonic) \
    INSTR2_INTEL(_id, _mnemonic)

#define INSTR3(_id, _mnemonic) \
    INSTR3_ATT(_id, _mnemonic) \
    INSTR3_INTEL(_id, _mnemonic)

#define DEFINE_OPCODE0_(_opcode, _mnemonic) INSTR0(_opcode, _mnemonic)
#define DEFINE_OPCODE0_REPEATABLE(_opcode, _mnemonic) \
    INSTR_REPEATED0(intel_##_opcode, _mnemonic) INSTR_REPEATED0(att_##_opcode, _mnemonic)
#define DEFINE_OPCODE0_PREFIX(_opcode, _mnemonic) \
    INSTR_PREFIX0(_opcode, intel, _mnemonic) INSTR_PREFIX0(_opcode, att, _mnemonic)
#define DEFINE_OPCODE0(_opcode, _mnemonic, _variant, _flags) DEFINE_OPCODE0_##_variant(_opcode, _mnemonic)
#define DEFINE_OPCODE1_(_opcode, _mnemonic) INSTR1(_opcode, _mnemonic)
#define DEFINE_OPCODE1_BRANCH(_opcode, _mnemonic) INSTR1_INTEL(_opcode, _mnemonic) INSTR1_ATT_BR(_opcode, _mnemonic)
#define DEFINE_OPCODE1(_opcode, _mnemonic, _variant, _flags, _op1) DEFINE_OPCODE1_##_variant(_opcode, _mnemonic)
#define DEFINE_OPCODE2(_opcode, _mnemonic, _variant, _flags, _op1, _op2) INSTR2(_opcode, _mnemonic)
#define DEFINE_OPCODE3(_opcode, _mnemonic, _variant, _flags, _op1, _op2, _op3) INSTR3(_opcode, _mnemonic)
KEFIR_AMD64_INSTRUCTION_DATABASE(DEFINE_OPCODE0, DEFINE_OPCODE1, DEFINE_OPCODE2, DEFINE_OPCODE3, )

#undef DEFINE_OPCODE0_
#undef DEFINE_OPCODE0_REPEATABLE
#undef DEFINE_OPCODE0_PREFIX
#undef DEFINE_OPCODE0
#undef DEFINE_OPCODE1_
#undef DEFINE_OPCODE1_BRANCH
#undef DEFINE_OPCODE1
#undef DEFINE_OPCODE2
#undef DEFINE_OPCODE3

#undef INSTR0
#undef INSTR1
#undef INSTR2
#undef INSTR3
#undef INSTR_REPEATED0

#undef INSTR0_ATT
#undef INSTR1_ATT
#undef INSTR1_ATT_BR
#undef INSTR2_ATT
#undef INSTR3_ATT

#undef INSTR0_INTEL
#undef INSTR1_INTEL
#undef INSTR1_INTEL
#undef INSTR2_INTEL
#undef INSTR3_INTEL

struct xasmgen_debug_info_tracker {
    struct kefir_hashtree files;
    kefir_uint64_t next_fileno;
};

static kefir_result_t free_debug_info_tracker_file(struct kefir_mem *mem, struct kefir_hashtree *tree,
                                                   kefir_hashtree_key_t key, kefir_hashtree_value_t value,
                                                   void *payload) {
    UNUSED(tree);
    UNUSED(value);
    UNUSED(payload);
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    ASSIGN_DECL_CAST(char *, file_name, key);
    REQUIRE(file_name != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid debug info file name"));

    KEFIR_FREE(mem, file_name);
    return KEFIR_OK;
}

static kefir_result_t new_debug_info_tracker(struct kefir_mem *mem, struct kefir_amd64_xasmgen *xasmgen,
                                             kefir_amd64_xasmgen_debug_info_tracker_t *tracker_ptr) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid amd64 xasmgen"));
    REQUIRE(tracker_ptr != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to amd64 xasmgen debug info tracker"));

    struct xasmgen_debug_info_tracker *tracker = KEFIR_MALLOC(mem, sizeof(struct xasmgen_debug_info_tracker));
    REQUIRE(tracker != NULL,
            KEFIR_SET_ERROR(KEFIR_MEMALLOC_FAILURE, "Failed to allocate amd64 xasmgen debug info tracker"));

    tracker->next_fileno = 0;
    kefir_result_t res = kefir_hashtree_init(&tracker->files, &kefir_hashtree_str_ops);
    REQUIRE_CHAIN(&res, kefir_hashtree_on_removal(&tracker->files, free_debug_info_tracker_file, NULL));
    REQUIRE_ELSE(res == KEFIR_OK, {
        KEFIR_FREE(mem, tracker);
        return res;
    });

    *tracker_ptr = (kefir_amd64_xasmgen_debug_info_tracker_t) tracker;
    return KEFIR_OK;
}

static kefir_result_t free_debug_info_tracker(struct kefir_mem *mem, struct kefir_amd64_xasmgen *xasmgen,
                                              kefir_amd64_xasmgen_debug_info_tracker_t tracker) {
    UNUSED(xasmgen);
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    ASSIGN_DECL_CAST(struct xasmgen_debug_info_tracker *, debug_info_tracker, tracker);
    REQUIRE(debug_info_tracker != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid amd64 xasmgen debug info tracker"));

    REQUIRE_OK(kefir_hashtree_free(mem, &debug_info_tracker->files));
    KEFIR_FREE(mem, debug_info_tracker);
    return KEFIR_OK;
}

static kefir_result_t debug_info_source_location(struct kefir_mem *mem, struct kefir_amd64_xasmgen *xasmgen,
                                                 kefir_amd64_xasmgen_debug_info_tracker_t tracker,
                                                 const struct kefir_source_location *source_location) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid amd64 xasmgen"));
    ASSIGN_DECL_CAST(struct xasmgen_debug_info_tracker *, debug_info_tracker, tracker);
    REQUIRE(debug_info_tracker != NULL,
            KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid amd64 xasmgen debug info tracker"));
    REQUIRE(source_location != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid source location"));

    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);

    switch (payload->syntax) {
        case KEFIR_AMD64_XASMGEN_SYNTAX_ATT:
        case KEFIR_AMD64_XASMGEN_SYNTAX_INTEL_NOPREFIX:
        case KEFIR_AMD64_XASMGEN_SYNTAX_INTEL_PREFIX: {
            struct kefir_hashtree_node *node;
            kefir_result_t res =
                kefir_hashtree_at(&debug_info_tracker->files, (kefir_hashtree_key_t) source_location->source, &node);
            uint64_t fileno;
            if (res == KEFIR_NOT_FOUND) {
                const kefir_size_t source_id_length = strlen(source_location->source);
                char *source_id = KEFIR_MALLOC(mem, (source_id_length + 1) * sizeof(char));
                REQUIRE(source_id != NULL,
                        KEFIR_SET_ERROR(KEFIR_MEMALLOC_FAILURE, "Failed to allocate source identifier"));
                strcpy(source_id, source_location->source);
                fileno = debug_info_tracker->next_fileno;

                res = kefir_hashtree_insert(mem, &debug_info_tracker->files, (kefir_hashtree_key_t) source_id,
                                            (kefir_hashtree_value_t) fileno);
                REQUIRE_ELSE(res == KEFIR_OK, {
                    KEFIR_FREE(mem, source_id);
                    return KEFIR_OK;
                });
                debug_info_tracker->next_fileno++;

                fprintf(payload->output, ".file %" KEFIR_UINT64_FMT " ", fileno);
                REQUIRE_OK(amd64_string_literal(print_fprintf, payload->output, source_id, source_id_length));
                fprintf(payload->output, "\n");
            } else {
                REQUIRE_OK(res);
                fileno = (kefir_uint64_t) node->value;
            }

            fprintf(payload->output, ".loc %" KEFIR_UINT64_FMT " %" KEFIR_UINT64_FMT " %" KEFIR_UINT64_FMT "\n", fileno,
                    (kefir_uint64_t) source_location->line, (kefir_uint64_t) source_location->column);
        } break;

        case KEFIR_AMD64_XASMGEN_SYNTAX_YASM:
            return KEFIR_SET_ERROR(KEFIR_NOT_SUPPORTED,
                                   "Debug information generation is not supported for Yasm backend");
    }
    return KEFIR_OK;
}

kefir_result_t kefir_asm_amd64_xasmgen_init(struct kefir_mem *mem, struct kefir_amd64_xasmgen *xasmgen, FILE *output,
                                            kefir_asm_amd64_xasmgen_syntax_t syntax) {
    REQUIRE(mem != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid memory allocator"));
    REQUIRE(xasmgen != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to amd64 xasmgen"));
    REQUIRE(output != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid output FILE"));

    struct xasmgen_payload *payload = KEFIR_MALLOC(mem, sizeof(struct xasmgen_payload));
    REQUIRE(payload != NULL, KEFIR_SET_ERROR(KEFIR_MEMALLOC_FAILURE, "Failed to allocator amd64 xasmgen payload"));
    payload->output = output;
    payload->syntax = syntax;

    xasmgen->payload = payload;

    xasmgen->settings.enable_comments = true;
    xasmgen->settings.enable_identation = true;

    xasmgen->prologue = amd64_prologue;
    xasmgen->close = amd64_close;
    xasmgen->newline = amd64_newline;
    xasmgen->comment = amd64_comment;
    if (payload->syntax == KEFIR_AMD64_XASMGEN_SYNTAX_YASM) {
        xasmgen->label = amd64_yasm_label;
        xasmgen->global = amd64_yasm_global;
        xasmgen->external = amd64_yasm_external;
        xasmgen->alias = amd64_yasm_alias;
        xasmgen->weak = amd64_yasm_weak;
        xasmgen->section = amd64_yasm_section;
        xasmgen->align = amd64_yasm_align;
        xasmgen->data = amd64_yasm_data;
        xasmgen->zerodata = amd64_yasm_zerodata;
        xasmgen->uninitdata = amd64_yasm_uninitdata;
        xasmgen->bindata = amd64_yasm_bindata;
    } else {
        xasmgen->label = amd64_label;
        xasmgen->global = amd64_global;
        xasmgen->external = amd64_external;
        xasmgen->alias = amd64_alias;
        xasmgen->weak = amd64_weak;
        xasmgen->section = amd64_section;
        xasmgen->align = amd64_align;
        xasmgen->data = amd64_data;
        xasmgen->zerodata = amd64_zerodata;
        xasmgen->uninitdata = amd64_uninitdata;
        xasmgen->bindata = amd64_bindata;
    }
    xasmgen->inline_assembly = amd64_inline_assembly;
    xasmgen->format_operand = amd64_format_operand;
    xasmgen->new_debug_info_tracker = new_debug_info_tracker;
    xasmgen->free_debug_info_tracker = free_debug_info_tracker;
    xasmgen->debug_info_source_location = debug_info_source_location;

    if (syntax != KEFIR_AMD64_XASMGEN_SYNTAX_ATT) {
#define OPCODE_DEF(_id) xasmgen->instr._id = amd64_instr_intel_##_id
#define DEFINE_OPCODE0(_opcode, _mnemonic, _variant, _flags) OPCODE_DEF(_opcode)
#define DEFINE_OPCODE1(_opcode, _mnemonic, _variant, _flags, _op1) OPCODE_DEF(_opcode)
#define DEFINE_OPCODE2(_opcode, _mnemonic, _variant, _flags, _op1, _op2) OPCODE_DEF(_opcode)
#define DEFINE_OPCODE3(_opcode, _mnemonic, _variant, _flags, _op1, _op2, _op3) OPCODE_DEF(_opcode)
        KEFIR_AMD64_INSTRUCTION_DATABASE(DEFINE_OPCODE0, DEFINE_OPCODE1, DEFINE_OPCODE2, DEFINE_OPCODE3, ;);
#undef OPCODE_DEF
    } else {
#define OPCODE_DEF(_id) xasmgen->instr._id = amd64_instr_att_##_id
        KEFIR_AMD64_INSTRUCTION_DATABASE(DEFINE_OPCODE0, DEFINE_OPCODE1, DEFINE_OPCODE2, DEFINE_OPCODE3, ;);
#undef OPCODE_DEF
#undef DEFINE_OPCODE0
#undef DEFINE_OPCODE1
#undef DEFINE_OPCODE2
    }
    return KEFIR_OK;
}

FILE *kefir_asm_amd64_xasmgen_get_output(const struct kefir_amd64_xasmgen *xasmgen) {
    REQUIRE(xasmgen != NULL, NULL);
    ASSIGN_DECL_CAST(struct xasmgen_payload *, payload, xasmgen->payload);
    REQUIRE(payload != NULL, NULL);

    return payload->output;
}

kefir_result_t kefir_asm_amd64_xasmgen_line_comment_prefix(const struct kefir_amd64_xasmgen *xasmgen,
                                                           const char **prefix_ptr) {
    UNUSED(xasmgen);
    REQUIRE(prefix_ptr != NULL, KEFIR_SET_ERROR(KEFIR_INVALID_PARAMETER, "Expected valid pointer to string"));

    *prefix_ptr = "#";
    return KEFIR_OK;
}

const struct kefir_asm_amd64_xasmgen_operand operand_regs[] = {
#define REG(x) [x] = {.klass = KEFIR_AMD64_XASMGEN_OPERAND_REGISTER, .reg = x}
    REG(KEFIR_AMD64_XASMGEN_REGISTER_AL),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_AH),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_BL),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_BH),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_CL),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_CH),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_DL),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_DH),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_SIL),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_DIL),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_SPL),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_BPL),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R8B),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R9B),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R10B),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R11B),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R12B),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R13B),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R14B),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R15B),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_AX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_BX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_CX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_DX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_SI),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_DI),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_SP),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_BP),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R8W),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R9W),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R10W),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R11W),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R12W),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R13W),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R14W),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R15W),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_EAX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_EBX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_ECX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_EDX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_ESI),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_EDI),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_ESP),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_EBP),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R8D),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R9D),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R10D),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R11D),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R12D),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R13D),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R14D),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R15D),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_RAX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_RBX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_RCX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_RDX),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_RSI),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_RDI),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_RSP),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_RBP),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R8),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R9),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R10),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R11),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R12),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R13),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R14),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_R15),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM0),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM1),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM2),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM3),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM4),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM5),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM6),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM7),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM8),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM9),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM10),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM11),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM12),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM13),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM14),
    REG(KEFIR_AMD64_XASMGEN_REGISTER_XMM15)
#undef REG
};

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_reg(
    kefir_asm_amd64_xasmgen_register_t reg) {
    return &operand_regs[reg];
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_imm(
    struct kefir_asm_amd64_xasmgen_operand *op, kefir_int64_t imm) {
    REQUIRE(op != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_IMMEDIATE;
    op->imm = imm;
    return op;
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_immu(
    struct kefir_asm_amd64_xasmgen_operand *op, kefir_uint64_t immu) {
    REQUIRE(op != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_IMMEDIATE_UNSIGNED;
    op->immu = immu;
    return op;
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_label(
    struct kefir_asm_amd64_xasmgen_operand *op, kefir_asm_amd64_xasmgen_symbol_relocation_t type, const char *label) {
    REQUIRE(op != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_LABEL;
    op->label.symbol = label;
    op->label.type = type;
    return op;
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_indirect(
    struct kefir_asm_amd64_xasmgen_operand *op, const struct kefir_asm_amd64_xasmgen_operand *base,
    kefir_int64_t disp) {
    REQUIRE(op != NULL, NULL);
    REQUIRE(base != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_INDIRECTION;
    op->indirection.base = base;
    op->indirection.displacement = disp;
    return op;
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_offset(
    struct kefir_asm_amd64_xasmgen_operand *op, const struct kefir_asm_amd64_xasmgen_operand *base,
    kefir_int64_t offset) {
    REQUIRE(op != NULL, NULL);
    REQUIRE(base != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_OFFSET;
    op->offset.base = base;
    op->offset.offset = offset;
    return op;
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_segment(
    struct kefir_asm_amd64_xasmgen_operand *op, kefir_asm_amd64_xasmgen_segment_register_t segment,
    const struct kefir_asm_amd64_xasmgen_operand *base) {
    REQUIRE(op != NULL, NULL);
    REQUIRE(base != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_SEGMENT;
    op->segment.segment = segment;
    op->segment.base = base;
    return op;
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_pointer(
    struct kefir_asm_amd64_xasmgen_operand *op, kefir_asm_amd64_xasmgen_pointer_type_t type,
    const struct kefir_asm_amd64_xasmgen_operand *base) {
    REQUIRE(op != NULL, NULL);
    REQUIRE(base != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_POINTER;
    op->pointer.type = type;
    op->pointer.base = base;
    return op;
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_rip_indirection(
    struct kefir_asm_amd64_xasmgen_operand *op, kefir_asm_amd64_xasmgen_symbol_relocation_t type,
    const char *identifier) {
    REQUIRE(op != NULL, NULL);
    REQUIRE(identifier != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_RIP_INDIRECTION;
    op->label.symbol = identifier;
    op->label.type = type;
    return op;
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_string_literal(
    struct kefir_asm_amd64_xasmgen_operand *op, const char *content, kefir_size_t length) {
    REQUIRE(op != NULL, NULL);
    REQUIRE(content != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_STRING_LITERAL;
    op->string_literal.content = content;
    op->string_literal.length = length;
    return op;
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_fpu_register(
    struct kefir_asm_amd64_xasmgen_operand *op, kefir_uint64_t reg) {
    REQUIRE(op != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_FPU_REGISTER;
    op->fpu_register = reg;
    return op;
}

const struct kefir_asm_amd64_xasmgen_operand *kefir_asm_amd64_xasmgen_operand_subtract(
    struct kefir_asm_amd64_xasmgen_operand *op, const struct kefir_asm_amd64_xasmgen_operand *left,
    const struct kefir_asm_amd64_xasmgen_operand *right) {
    REQUIRE(op != NULL, NULL);
    op->klass = KEFIR_AMD64_XASMGEN_OPERAND_SUBTRACT;
    op->arithmetics.args[0] = left;
    op->arithmetics.args[1] = right;
    return op;
}

const char *kefir_asm_amd64_xasmgen_helpers_format(struct kefir_asm_amd64_xasmgen_helpers *helpers, const char *fmt,
                                                   ...) {
    REQUIRE(helpers != NULL, NULL);

    va_list args;
    va_start(args, fmt);
    vsnprintf(helpers->buffer, KEFIR_AMD64_XASMGEN_HELPERS_BUFFER_LENGTH - 1, fmt, args);
    va_end(args);
    return helpers->buffer;
}
