/*
    SPDX-License-Identifier: GPL-3.0

    Copyright (C) 2020-2025  Jevgenijs Protopopovs

    This file is part of Kefir project.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "kefir/core/mem.h"
#include "kefir/parser/parser.h"
#include "kefir/ast/format.h"
#include <stdio.h>

kefir_result_t kefir_int_test(struct kefir_mem *mem) {
    struct kefir_string_pool symbols;
    struct kefir_token TOKENS[1024];
    struct kefir_parser_token_cursor cursor;
    struct kefir_parser parser;

    REQUIRE_OK(kefir_string_pool_init(&symbols));

    kefir_size_t counter = 0;

    const char Literal1[] = "Literal1 string literal. It's a normal one";
    const char Literal2[] = u8"Literal2 string literal. It's a UTF-8 literal. Similar to normal one";
    const kefir_char16_t Literal3[] = u"Crazy 2-byte string literal. It has char16_t type";
    const kefir_char32_t Literal4[] = U"Even more crazy string literal. It's represented as int";
    const kefir_wchar_t Literal5[] = L"Wide string literal. Not so crazy though";

    REQUIRE_OK(kefir_token_new_string_literal_multibyte(mem, Literal1, sizeof(Literal1), &TOKENS[counter++]));
    REQUIRE_OK(kefir_token_new_string_literal_unicode8(mem, Literal2, sizeof(Literal2), &TOKENS[counter++]));
    REQUIRE_OK(kefir_token_new_string_literal_unicode16(mem, Literal3, sizeof(Literal3) / sizeof(Literal3[0]),
                                                        &TOKENS[counter++]));
    REQUIRE_OK(kefir_token_new_string_literal_unicode32(mem, Literal4, sizeof(Literal4) / sizeof(Literal4[0]),
                                                        &TOKENS[counter++]));
    REQUIRE_OK(
        kefir_token_new_string_literal_wide(mem, Literal5, sizeof(Literal5) / sizeof(Literal5[0]), &TOKENS[counter++]));

    REQUIRE_OK(kefir_parser_token_cursor_init_direct(&cursor, TOKENS, counter));
    REQUIRE_OK(kefir_parser_init(mem, &parser, &symbols, &cursor, NULL));

    struct kefir_json_output json;
    REQUIRE_OK(kefir_json_output_init(&json, stdout, 4));

    REQUIRE_OK(kefir_json_output_array_begin(&json));

    while (kefir_parser_token_cursor_at(&cursor, 0)->klass != KEFIR_TOKEN_SENTINEL) {
        struct kefir_ast_node_base *node = NULL;
        REQUIRE_OK(KEFIR_PARSER_NEXT_EXPRESSION(mem, &parser, &node));
        REQUIRE_OK(kefir_ast_format(&json, node, false));
        REQUIRE_OK(KEFIR_AST_NODE_FREE(mem, node));
    }

    REQUIRE_OK(kefir_json_output_array_end(&json));
    REQUIRE_OK(kefir_json_output_finalize(&json));

    REQUIRE_OK(kefir_parser_free(mem, &parser));
    for (kefir_size_t i = 0; i < counter; i++) {
        REQUIRE_OK(kefir_token_free(mem, &TOKENS[i]));
    }
    REQUIRE_OK(kefir_string_pool_free(mem, &symbols));
    return KEFIR_OK;
}
